/*
	$Id $

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.AnnotationTypeCategory;
import net.sf.basedb.core.BioPlateType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.formatter.Formatter;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

/**
	Some utility functions for annotations.
	@author nicklas
	@since 3.5
*/
public class AnnotationUtil 
{

	/**
		Load information about categories and subtypes of an annotation type that
		are useful for deciding about inheriting annotations from parent items,
		The result is returned as a JSON object intended to be used by the
		"Batch inherit annotations" feature in the web interface. 
		See: common/annotations/batch_inherit.jsp (and .js)
	*/
	@SuppressWarnings("unchecked")
	public static JSONObject loadJsonForBatchInherit(DbControl dc, AnnotationType at)
	{
		JSONObject json = new JSONObject();
		json.put("id", at.getId());
		json.put("name", at.getName());
		if (at.getDisableInheritance())
		{
			json.put("disableInheritance", 1);
			return json;
		}
		// Load categories for the annotation type
		ItemQuery<AnnotationTypeCategory> categoryQuery = AnnotationTypeCategory.getQuery();
		categoryQuery.include(Include.MINE, Include.OTHERS, Include.SHARED, Include.IN_PROJECT);
		categoryQuery.join(Hql.innerJoin("annotationTypes", "atp"));
		categoryQuery.restrict(Restrictions.eq(Hql.alias("atp"), Hql.entity(at)));
		categoryQuery.order(Orders.asc(Hql.property("name")));
		
		JSONArray jsonCategories = new JSONArray();
		Set<String> names = new HashSet<String>();
		for (AnnotationTypeCategory cat : categoryQuery.list(dc))
		{
			JSONObject jsonCat = new JSONObject();
			jsonCat.put("id", cat.getId());
			jsonCat.put("name", cat.getName());
			names.add(cat.getName());
			jsonCategories.add(jsonCat);
		}
		json.put("categories", jsonCategories);
		
		// Load subtypes for item types matching items types from the annotation type
		ItemQuery<ItemSubtype> subtypeQuery = ItemSubtype.getQuery(null);
		subtypeQuery.order(Orders.asc(Hql.property("name")));
		subtypeQuery.restrict(Restrictions.in(Hql.property("itemType"), Expressions.parameter("itemTypes")));
		Set<Integer> itemTypes = new HashSet<Integer>();
		boolean hasBioPlateAnnotation = false;
		for (Item itemType : at.getEnabledItems())
		{
			if (itemType == Item.BIOPLATE) hasBioPlateAnnotation = true;
			itemTypes.add(itemType.getValue());
		}
		subtypeQuery.setParameter("itemTypes", itemTypes, Type.INT);
		JSONArray jsonSubtypes = new JSONArray();
		boolean hasDefault = false;
		for (ItemSubtype subtype : subtypeQuery.list(dc))
		{
			JSONObject jsonSubtype = new JSONObject();
			jsonSubtype.put("id", subtype.getId());
			jsonSubtype.put("name", subtype.getName());
			jsonSubtype.put("itemType", subtype.getMainItemType().name());
			if (!hasDefault && names.contains(subtype.getName()))
			{
				jsonSubtype.put("selected", 1);
				hasDefault = true;
			}
			jsonSubtypes.add(jsonSubtype);
		}
		json.put("subtypes", jsonSubtypes);
		
		if (hasBioPlateAnnotation && names.size() > 0)
		{
			// Load bioplate types linked with subtypes having the same name as the categories
			ItemQuery<BioPlateType> bioPlateTypeQuery = BioPlateType.getQuery();
			bioPlateTypeQuery.join(Hql.innerJoin("itemSubtype", "st"));
			bioPlateTypeQuery.restrict(Restrictions.in(Hql.property("st", "name"), Expressions.parameter("categories")));
			bioPlateTypeQuery.order(Orders.asc(Hql.property("name")));
			bioPlateTypeQuery.setParameter("categories", names, Type.STRING);
			
			JSONArray jsonPlateTypes = new JSONArray();
			for (BioPlateType plateType : bioPlateTypeQuery.list(dc))
			{
				ItemSubtype subtype = plateType.getItemSubtype();
				
				JSONObject jsonPlateType = new JSONObject();
				jsonPlateType.put("id", plateType.getId());
				jsonPlateType.put("name", plateType.getName());
				
				JSONObject jsonSubtype = new JSONObject();
				jsonSubtype.put("id", subtype.getId());
				jsonSubtype.put("name", subtype.getName());
				jsonSubtype.put("itemType", subtype.getMainItemType().name());
				jsonPlateType.put("subtype", jsonSubtype);

				jsonPlateTypes.add(jsonPlateType);
			}
			json.put("plateTypes", jsonPlateTypes);
		}

		return json;
	}
	
	
	@SuppressWarnings("unchecked")
	public static JSONObject createJsonForAnnotationTypeAndAnnotation(DbControl dc, AnnotationType at, 
		Annotation a, Annotation inherited, boolean isProtocolParameter, Set<AnnotationTypeCategory> allCategories)
	{
		// Create the JSON objects for holding all info
		// The base object has 'id', 'source', 'annotationType' and 'annotation'
		JSONObject json = new JSONObject();
		String id = ""+at.getId();
		if (a != null) id += "-" + a.getId();
		if (inherited != null) id += "-" + inherited.getId();
		json.put("id", id);
		// If 'a' is null this is a primary annotation that has no current values
		json.put("source", a == null ? (inherited == null ? "PRIMARY" : "INHERITED") : a.getSource().name());

		// Flag indicating if the user has permission to modify the annotation
		json.put("annotatePermission", a != null ? a.hasPermission(Permission.WRITE) : at.hasPermission(Permission.USE));
		
		JSONObject jsonAt = new JSONObject();
		json.put("annotationType", jsonAt);
		JSONObject jsonA = new JSONObject();
		json.put("annotation", jsonA);
		JSONArray jsonValues = new JSONArray();
		jsonA.put("values", jsonValues);

		// Load annotation type information
		Type valueType = at.getValueType();
		Formatter<Object> formatter = (Formatter<Object>)FormatterFactory.getTypeFormatter(dc.getSessionControl(), valueType);
		
		jsonAt.put("id", at.getId());
		jsonAt.put("name", at.getName());
		jsonAt.put("valueType", valueType.name());
		jsonAt.put("multiplicity", at.getMultiplicity());
		jsonAt.put("description", HTML.encodeTags(at.getDescription()));
		jsonAt.put("removed", at.isRemoved());
		jsonAt.put("protocolParameter", isProtocolParameter);
		jsonAt.put("projectAnnotations", at.getProjectAnnotations());

		if (valueType == Type.STRING)
		{
			jsonAt.put("maxLength", at.getMaxLength());
		}
		else if (valueType == Type.INT || valueType == Type.LONG)
		{
			jsonAt.put("minValue", at.getMinValueLong());
			jsonAt.put("maxValue", at.getMaxValueLong());
		}
		else if (valueType == Type.FLOAT || valueType == Type.DOUBLE)
		{
			jsonAt.put("minValue", at.getMinValueDouble());
			jsonAt.put("maxValue", at.getMaxValueDouble());
		}
		
		// Enumeration
		if (at.isEnumeration())
		{
			List<?> values = at.getValues();
			JSONArray jsonEnum = new JSONArray();
			for (Object val : values)
			{
				jsonEnum.add(formatter.format(val));
			}
			jsonAt.put("enumeration", jsonEnum);
			jsonAt.put("displayAsList", at.getDisplayAsList());
		}
		
		// Units
		if (at.supportUnits())
		{
			Unit defaultUnit = at.getDefaultUnit();
			jsonAt.put("defaultUnit", defaultUnit.getId());
			jsonAt.put("defaultSymbol", defaultUnit.getDisplaySymbol());
			
			if (!at.isEnumeration())
			{
				ItemQuery<Unit> unitQuery = at.getUsableUnits();
				if (unitQuery.count(dc) == 0)
				{
					unitQuery = at.getQuantity().getUnits();
				}
				else
				{
					unitQuery.reset();
				}
				unitQuery.order(Orders.desc(Hql.property("referenceFactor")));
				unitQuery.order(Orders.asc(Hql.property("displaySymbol")));
				List<Unit> units = new ArrayList<Unit>(unitQuery.list(dc));
				if (!units.contains(defaultUnit)) units.add(defaultUnit);
				
				Unit currentUnit = a != null ? a.getUnit() : null;
				if (currentUnit != null && !units.contains(currentUnit)) units.add(currentUnit);
				
				JSONArray jsonUnits = new JSONArray();
				for (Unit unit : units)
				{
					JSONObject jsonUnit = new JSONObject();
					jsonUnit.put("id", unit.getId());
					jsonUnit.put("description", HTML.encodeTags(unit.getName() + " - " + unit.getDescription()));
					jsonUnit.put("symbol", HTML.encodeTags(unit.getDisplaySymbol()));
					jsonUnits.add(jsonUnit);
				}
				jsonAt.put("units", jsonUnits);
			}
		}
		
		// Categories
		JSONArray jsonCategories = new JSONArray();
		ItemQuery<AnnotationTypeCategory> categoryQuery = at.getCategories();
		categoryQuery.include(Include.MINE, Include.OTHERS, Include.SHARED, Include.IN_PROJECT);
		categoryQuery.join(Hql.innerJoin("annotationTypes", "atp"));
		categoryQuery.restrict(Restrictions.eq(Hql.alias("atp"), Hql.entity(at)));
		for (AnnotationTypeCategory category : categoryQuery.list(dc))
		{
			if (allCategories != null) allCategories.add(category);
			jsonCategories.add(category.getId());
		}
		jsonAt.put("categories", jsonCategories);

		
		if (a != null)
		{
			createJSON(jsonA, dc, a, formatter, false);
			
			if (a.getSource() != Annotation.Source.PRIMARY)
			{
				inherited = a.getInheritedFrom();
			}
			
			if (a.isOverridingDefault())
			{
				Annotation defA = a.getDefaultAnnotation();
				JSONObject jsonDefault = new JSONObject();
				createJSON(jsonDefault, dc, defA, formatter, true);
				json.put("defaultValues", jsonDefault);
			}
		}
		
		if (inherited != null)
		{
			JSONObject jsonInherited = new JSONObject();
			createJSON(jsonInherited, dc, inherited, formatter, true);
			Annotatable inheritedFrom = inherited.getAnnotationSet().getItem(dc);
			JSONObject jsonFrom = new JSONObject();
			jsonFrom.put("id", inheritedFrom.getId());
			jsonFrom.put("name", ((Nameable)inheritedFrom).getName());
			jsonFrom.put("type", inheritedFrom.getType().name());
			jsonInherited.put("from", jsonFrom);
			json.put("inherited", jsonInherited);
		}
		

		return json;
	}

	@SuppressWarnings("unchecked")
	static void createJSON(JSONObject json, DbControl dc, Annotation a, Formatter<Object> formatter, boolean formatAll)
	{
		json.put("id", a.getId());
		json.put("projectId", a.getProjectId());
		json.put("override", a.isOverridingDefault());
		Date lastUpdate = a.getLastUpdate();
		if (lastUpdate != null) json.put("lastUpdate",  lastUpdate.getTime());
		
		Unit currentUnit = a.getUnit();
		if (currentUnit != null) 
		{
			json.put("unit", currentUnit.getId());
			json.put("unitSymbol", currentUnit.getDisplaySymbol());
		}
		
		// Populate the values array
		List<?> values = a.getValues(null);
		if (values != null)
		{
			JSONArray jsonValues = new JSONArray();
			json.put("values", jsonValues);
			
			for (Object value : values)
			{
				// Dates should be formatted to strings using the current date(time) format
				if (formatAll || value instanceof Date) 
				{
					value = formatter.format(value);
				}
				jsonValues.add(value.toString());
			}
		}
		
	}

}
