/**
	$Id: ChangeHistory.java 7166 2016-06-07 06:20:23Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Table;

import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.ThreadSignalHandler;

/**

	@author Nicklas
	@version 2.13
	@base.modified $Date: 2016-06-07 08:20:23 +0200 (ti, 07 jun 2016) $
*/
public class ChangeHistory
	extends BasicItem
{

	/**
		The type of item represented by this class.
		@see Item#CHANGEHISTORY
		@see #getType()
	*/
	public static final Item TYPE = Item.CHANGEHISTORY;

	/**
		Get a <code>ChangeHistory</code> item when you know the ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>ChangeHistory</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static ChangeHistory getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ChangeHistory ch = dc.loadItem(ChangeHistory.class, id);
		if (ch == null) throw new ItemNotFoundException("ChangeHistory[id="+id+"]");
		return ch;
	}
	
	/**
		Get the history for a given item.
		@param item The item to get the history for
		@throws InvalidUseOfNullException If the item is null
	*/
	public static ItemQuery<ChangeHistory> getHistoryOf(BasicItem item)
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		ItemQuery<ChangeHistory> query = new ItemQuery<ChangeHistory>(ChangeHistory.class);
		query.joinPermanent(
				Hql.innerJoin(null, "changeHistory", "hst", true)
		);
		query.restrictPermanent(
			Restrictions.and(
				Restrictions.eq(
						Hql.property("itemId"), 
						Expressions.integer(item.getId())),
				Restrictions.eq(
						Hql.property("itemType"), 
						Expressions.integer(item.getType().getValue()))
			));
		return query;
	}
	
	/**
		Get all changes made by the specified user.
		@param user A user object
		@throws InvalidUseOfNullException If the user is null
	*/
	public static ItemQuery<ChangeHistory> getChangesBy(User user)
	{
		if (user == null) throw new InvalidUseOfNullException("user");
		ItemQuery<ChangeHistory> query = new ItemQuery<ChangeHistory>(ChangeHistory.class);
		query.joinPermanent(
				Hql.innerJoin(null, "changeHistory", "hst", true)
		);
		query.restrictPermanent(
				Restrictions.eq(
						Hql.property("hst", "userId"), 
						Expressions.integer(user.getId()))
			);
		return query;		
	}
	
	/**
		Get all changes made by the specified job.
		@param job A job object
		@throws InvalidUseOfNullException If the job is null
	*/
	public static ItemQuery<ChangeHistory> getChangesBy(Job job)
	{
		if (job == null) throw new InvalidUseOfNullException("job");
		ItemQuery<ChangeHistory> query = new ItemQuery<ChangeHistory>(ChangeHistory.class);
		query.joinPermanent(
				Hql.innerJoin(null, "changeHistory", "hst", true)
		);
		query.restrictPermanent(
				Restrictions.eq(
						Hql.property("hst", "jobId"), 
						Expressions.integer(job.getId()))
			);
		return query;		
	}
	
	/**
		Get all changes made in the specified session.
		@param session A session object
		@throws InvalidUseOfNullException If the session is null
	*/
	public static ItemQuery<ChangeHistory> getChangesIn(Session session)
	{
		if (session == null) throw new InvalidUseOfNullException("session");
		ItemQuery<ChangeHistory> query = new ItemQuery<ChangeHistory>(ChangeHistory.class);
		query.joinPermanent(
				Hql.innerJoin(null, "changeHistory", "hst", true)
		);
		query.restrictPermanent(
				Restrictions.eq(
						Hql.property("hst", "sessionId"), 
						Expressions.integer(session.getId()))
			);
		return query;		
	}
		
	/**
		Delete all history entries that are linking to non-existing items. This method 
		is intended to be executed at regular intervals by a cleanup application.

		@param progress An optional progress reporter
		@since 2.13
	*/
	public static synchronized int deleteStrayEntries(ProgressReporter progress)
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		int numDeleted = 0;
	
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			// Find the used items types in the history
			if (progress != null) 
			{
				progress.display(5, "Finding item types with history...");
			}
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
					"FIND_USED_TYPES_IN_CHANGEHISTORY");
			/*
				SELECT DISTINCT ch.itemType 
				FROM ChangeHistoryDetailData ch
			*/
			List<Integer> itemTypes = HibernateUtil.loadList(Integer.class, query, null);
			
			// For each item type: find and delete entries which reference non-existing items
			int index = 0;
			int numItemTypes = itemTypes.size();
			int maxParameters = HibernateUtil.getDbEngine().getMaxParametersInQuery();
			org.hibernate.Query deleteQuery = 
				HibernateUtil.getPredefinedQuery(session, "DELETE_STRAY_CHANGEHISTORY");
			for (int itemType : itemTypes)
			{
				ThreadSignalHandler.checkInterrupted();
				Item item = Item.fromValue(itemType);
				if (progress != null)
				{
					int percent = 10 + (90 * index) / numItemTypes;
					progress.display(percent, "Deleting history entries for non-existing item: " + item);
				}
				PersistentClass pClass = HibernateUtil.getClassMapping(item.getDataClass().getName());
				Table table = pClass.getTable();
			
				query = HibernateUtil.getPredefinedSQLQuery(session, "SELECT_STRAY_CHANGEHISTORY", 
						table.getName());
				/*
					SELECT [ch].[id]
					FROM [ChangeHistoryDetails] [ch] 
					LEFT JOIN [{1}] [t] 
						ON ([t].[id] = [ch].[item_id] AND [ch].[item_type] = :type)
					WHERE [t].[id] IS NULL 
					AND [ch].[item_type] = :type
				*/
				query.setInteger("type", itemType);
				List<Integer> stray = new ArrayList<Integer>(HibernateUtil.loadList(Integer.class, query, null));
				int startIndex = 0;
				int endIndex = Math.min(maxParameters, stray.size());
				while (startIndex < stray.size())
				{
					deleteQuery.setParameterList("ids", stray.subList(startIndex, endIndex), TypeWrapper.INTEGER.getHibernateType());
					numDeleted += HibernateUtil.executeUpdate(deleteQuery);
					startIndex = endIndex;
					endIndex = Math.min(startIndex + maxParameters, stray.size());
				}
				index++;
			}
			HibernateUtil.commit(tx);
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return numDeleted;
	}
	
	private boolean valuesPermissions;
	
	ChangeHistory(ChangeHistoryDetailData data)
	{
		super(data);
	}
	
	@Override
	ChangeHistoryDetailData getData()
	{
		return (ChangeHistoryDetailData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		WRITE permission and higher is always denied. READ permission
		is granted based on role permissions.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		valuesPermissions = Permission.hasPermission(getSessionControl().getRolePermissions(Item.CHANGEHISTORY_VALUES), Permission.READ);
		denied |= Permission.deny(Permission.WRITE);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Get the date and time the change was made.
	*/
	public Date getTime()
	{
		return DateUtil.copy(getData().getChangeHistory().getTime());
	}
	
	/**
		Get the type of change that was made.
	*/
	public ChangeType getChangeType()
	{
		return ChangeType.fromValue(getData().getChangeType());
	}
	
	/**
		Get the ID of the item that was changed.
	*/
	public int getItemId()
	{
		return getData().getItemId();
	}
	
	/**
		Get the type of the item that was changed.
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}
	
	/**
		Load the item that was changed. 
		@param dc A DbControl to use for database access, or null to
			use the same DbControl this history entry belongs to
		@return A BasicItem
		@throws ItemNotFoundException If the item no longer exists in
			the database
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to access the item
	*/
	public BasicItem getItem(DbControl dc)
	{
		if (dc == null) dc = getDbControl();
		return getItemType().getById(dc, getItemId());
	}
	
	/**
		Get some additional free-text information about the change.
		@return A string, or null if no information is available
	*/
	public String getChangeInfo()
	{
		return getData().getChangeInfo();
	}
	
	/**
		Get the old value of the property had before this change.
		@return The old value or null if not known or if the logged in
			user doesn't have permission to see it
		@since 3.3
	*/
	public String getOldValue()
	{
		return valuesPermissions ? getData().getOldValue() : null;
	}

	/**
		Get the new value of the property had immediately after this change.
		@return The new value or null if not known or if the logged in
			user doesn't have permission to see it
		@since 3.3
	*/
	public String getNewValue()
	{
		return valuesPermissions ? getData().getNewValue() : null;
	}
	
	/**
		Get the ID of the user that made the change.
	*/
	public int getUserId()
	{
		return getData().getChangeHistory().getUserId();
	}
	
	/**
		Load the user that made the change.
		@param dc A DbControl to use for database access, or null to
			use the same DbControl this history entry belongs to
		@throws ItemNotFoundException If the user no longer exists in
			the database
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to access the other user
	*/
	public User getUser(DbControl dc)
	{
		if (dc == null) dc = getDbControl();
		return User.getById(dc, getUserId());
	}
	
	/**
		Get the ID of the session in which the change was made.
	*/
	public int getSessionId()
	{
		return getData().getChangeHistory().getSessionId();
	}
	
	/**
		Load the session in which the change was made.
		@param dc A DbControl to use for database access, or null to
			use the same DbControl this history entry belongs to
		@throws ItemNotFoundException If the session no longer exists in
			the database
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to access the session
	*/
	public Session getSession(DbControl dc)
	{
		if (dc == null) dc = getDbControl();
		return Session.getById(dc, getSessionId());
	}

	/**
		Get the ID of the client application that was used to
		make the change.
		@return The ID or null if the client application is not known
	*/
	public Integer getClientId()
	{
		return getData().getChangeHistory().getClientId();
	}
	
	/**
		Load the client application that was used when the change was made.
		@param dc A DbControl to use for database access, or null to
			use the same DbControl this history entry belongs to
		@return A Client object or null if it is not known which client 
			that was used
		@throws ItemNotFoundException If the client application no longer exists in
			the database
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to access the client
	*/
	public Client getClient(DbControl dc)
	{
		if (dc == null) dc = getDbControl();
		Integer id = getClientId();
		return id == null ? null : Client.getById(dc, id);
	}

	/**
		Get the ID of the project that was active when the change
		was made.
		@return The ID or null if no project was active
	*/
	public Integer getProjectId()
	{
		return getData().getChangeHistory().getProjectId();
	}
	
	/**
		Load the project that was active when the change was made.
		@param dc A DbControl to use for database access, or null to
			use the same DbControl this history entry belongs to
		@return A Project object or null if no project was active
		@throws ItemNotFoundException If the client application no longer exists in
			the database
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to access the project
	*/
	public Project getProject(DbControl dc)
	{
		if (dc == null) dc = getDbControl();
		Integer id = getProjectId();
		return id == null ? null : Project.getById(dc, id);
	}

	/**
		Get the ID of the plug-in that was making the change.
		@return The ID or null if the change was not made by a plug-in
	*/
	public Integer getPluginId()
	{
		return getData().getChangeHistory().getPluginId();
	}
	
	/**
		Load the plug-in that made the change.
		@param dc A DbControl to use for database access, or null to
			use the same DbControl this history entry belongs to
		@return A PluginDefinition object or null if the change
			was not made by a plug-in
		@throws ItemNotFoundException If the plug-in no longer exists in
			the database
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to access the plug-in
	*/
	public PluginDefinition getPlugin(DbControl dc)
	{
		if (dc == null) dc = getDbControl();
		Integer id = getPluginId();
		return id == null ? null : PluginDefinition.getById(dc, id);
	}

	/**
		Get the ID of the job that was executing when the change was made
		@return The ID or null if the no job was executing
	*/
	public Integer getJobId()
	{
		return getData().getChangeHistory().getJobId();
	}
	
	/**
		Load the job that was executing when the change was made.
		@param dc A DbControl to use for database access, or null to
			use the same DbControl this history entry belongs to
		@return A Job object or null if no job was executing when the
			change was made
		@throws ItemNotFoundException If the job no longer exists in
			the database
		@throws PermissionDeniedException If the logged in user doesn't
			have permission to access the job
	*/
	public Job getJob(DbControl dc)
	{
		if (dc == null) dc = getDbControl();
		Integer id = getJobId();
		return id == null ? null : Job.getById(dc, id);
	}

}
