/**
	$Id: ClonableProperty.java 7052 2016-01-11 13:40:07Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.Formula.AverageMethod;
import net.sf.basedb.core.data.ReporterClonePropertyData;
import net.sf.basedb.core.data.ReporterData;

/**
	Represents a reporter property that can be cloned to a per-experiment
	table in the dynamic database. A cloned property is more or less the same
	as an {@link ExtendedProperty} but with added support for regular
	(static) properties.

	@author Nicklas
	@since 3.1
	@base.modified $Date: 2016-01-11 14:40:07 +0100 (må, 11 jan 2016) $
*/
public class ClonableProperty
	extends ExtendedProperty
	implements Comparable<ClonableProperty>
{
	
	private static final boolean MANDATORY = true;
	private static final boolean NULLABLE = true;
	
	/**
		The property name of the "static" property {@link ReporterData#getId()}.
		This property is mandatory in clone templates.
	*/
	public static String ID = "id";

	/**
		The property name of the "static" property {@link ReporterData#getVersion()}.
		This property is mandatory in clone templates.
	*/
	public static String VERSION = "version";

	/**
		The property name of the "static" property {@link ReporterData#getExternalId()}.
		This property is mandatory in clone templates.
	*/
	public static String EXTERNAL_ID = "externalId";

	/**
		The property name of the "static" property {@link ReporterData#getName()}
	*/
	public static String NAME = "name";

	/**
		The property name of the "static" property {@link ReporterData#getSymbol()}
	*/
	public static String SYMBOL = "symbol";

	/**
		The property name of the "static" property {@link ReporterData#getEntryDate()}
	*/
	public static String ENTRY_DATE = "entryDate";

	/**
		The property name of the "static" property {@link ReporterData#getLastSource()}
	*/
	public static String LAST_SOURCE = "lastSource";

	/**
		The property name of the "static" property {@link ReporterData#getLastUpdate()}
	*/
	public static String LAST_UPDATE = "lastUpdate";

	/**
		The property name of the "static" property {@link ReporterData#getDescription()}
	*/
	public static String DESCRIPTION = "description";

	/**
		Create a cloned property that represents the "static" property
		with the given name.
		@param name The name of the static property
		@return A cloned property
		@throws ItemNotFoundException If a property with the given name
			doesn't exist
	*/
	public static ClonableProperty createFromRegular(String name)
	{
		ClonableProperty property = null;
		if (ID.equals(name))
		{
			property = new ClonableProperty(ID, "Id", "The internal id of the reporter", "id", 
				Type.INT, 0, !NULLABLE, AverageMethod.NONE, MANDATORY);
		}
		else if (VERSION.equals(name))
		{
			property = new ClonableProperty(VERSION, "Version", "The internal version number of the reporter", "version", 
				Type.INT, 0, !NULLABLE, AverageMethod.NONE, MANDATORY);
		}
		else if (EXTERNAL_ID.equals(name))
		{
			property = new ClonableProperty(EXTERNAL_ID, "External ID", "The external id of the reporter", "external_id", 
				Type.STRING, ReporterData.MAX_EXTERNAL_ID_LENGTH, !NULLABLE, AverageMethod.NONE, MANDATORY);
		}
		else if (NAME.equals(name))
		{
			property = new ClonableProperty(NAME, "Name", "The name of the reporter", "name", 
				Type.STRING, ReporterData.MAX_NAME_LENGTH, !NULLABLE, AverageMethod.NONE, !MANDATORY);
		}
		else if (SYMBOL.equals(name))
		{
			property = new ClonableProperty(SYMBOL, "Gene symbol", "The gene symbol for the reporter", "symbol", 
				Type.STRING, ReporterData.MAX_SYMBOL_LENGTH, NULLABLE, AverageMethod.NONE, !MANDATORY);
		}
		else if (ENTRY_DATE.equals(name))
		{
			property = new ClonableProperty(ENTRY_DATE, "Entry date", "The date this reporter was registered in the database", "entry_date", 
				Type.DATE, 0, NULLABLE, AverageMethod.NONE, !MANDATORY);
		}
		else if (LAST_SOURCE.equals(name))
		{
			property = new ClonableProperty(LAST_SOURCE, "Last source", "The source (file) of the last update to the reporter", "last_source", 
				Type.STRING, ReporterData.MAX_LAST_SOURCE_LENGTH, NULLABLE, AverageMethod.NONE, !MANDATORY);
		}
		else if (LAST_UPDATE.equals(name))
		{
			property = new ClonableProperty(LAST_UPDATE, "Last update", "The date and time this reporter was last updated", "last_update", 
				Type.TIMESTAMP, 0, NULLABLE, AverageMethod.NONE, !MANDATORY);
		}
		else if (DESCRIPTION.equals(name))
		{
			property = new ClonableProperty(DESCRIPTION, "Description", "A custom description", "description", 
				Type.TEXT, ReporterData.MAX_DESCRIPTION_LENGTH, NULLABLE, AverageMethod.NONE, !MANDATORY);
		}
		else
		{
			throw new ItemNotFoundException("ClonedProperty[name="+name + "]");
		}
		return property;
	}
	
	/**
		Create a cloned property representation of an extended property.
		@param name The name of the extended property
		@return A cloned property view of the extended property
		@throws ItemNotFoundException If a property with the given name
			doesn't exist
	*/
	public static ClonableProperty createFromExtended(String name)
	{
		ExtendedProperty ep = ExtendedProperties.getProperty("ReporterData", name);
		if (ep == null)
		{
			throw new ItemNotFoundException("ExtendedProperty[name="+name+"]");
		}
		return new ClonableProperty(ep);
	}
	
	/**
		Create a cloned property from either an extended or static property.
		This method will first check if an extended property with the given
		name exists. 
		@param name The name of the property
		@return A cloned property view of the extended or static property
		@throws ItemNotFoundException If a property with the given name
			doesn't exist
	*/
	public static ClonableProperty create(String name)
	{
		ClonableProperty property = null;
		ExtendedProperty ep = ExtendedProperties.getProperty("ReporterData", name);
		if (ep == null)
		{
			property = createFromRegular(name);
		}
		else
		{
			property = new ClonableProperty(ep);
		}
		return property;
	}
	
	/**
		Get a list containing cloned property representation for all static and 
		extended properties.
	*/
	public static List<ClonableProperty> getAll()
	{
		List<ClonableProperty> all = getAllMandatory();
		// Other static properties
		all.add(createFromRegular(NAME));
		all.add(createFromRegular(SYMBOL));
		all.add(createFromRegular(DESCRIPTION));
		all.add(createFromRegular(ENTRY_DATE));
		all.add(createFromRegular(LAST_SOURCE));
		all.add(createFromRegular(LAST_UPDATE));
		// All extended properties
		List<ExtendedProperty> extended = ExtendedProperties.getProperties("ReporterData");
		if (extended != null)
		{
			for (ExtendedProperty ep : extended)
			{
				all.add(new ClonableProperty(ep));
			}
		}
		Collections.sort(all);
		return all;
	}
	
	/**
		Get a list with all mandatory properties that must be cloned.
	*/
	public static List<ClonableProperty> getAllMandatory()
	{
		List<ClonableProperty> mandatory = new ArrayList<ClonableProperty>();
		mandatory.add(createFromRegular(EXTERNAL_ID));
		mandatory.add(createFromRegular(ID));
		mandatory.add(createFromRegular(VERSION));
		return mandatory;
	}
	
	private final boolean isMandatory;
	private final boolean isExtendedProperty;
	
	/**
		Create a cloned property representation of a static property.
	*/
	private ClonableProperty(String propertyName, String title, String description,
			String column, Type valueType, int length, boolean nullable, AverageMethod averageMethod, boolean isMandatory)
	{
		super(propertyName, title, description, column, 
			valueType, length, nullable, true, true, 
			false, averageMethod, null);
		this.isExtendedProperty = false;
		this.isMandatory = isMandatory;
	}
	
	/**
		Create a cloned property representation of an extended property
		@param p The extended property to clone
	*/
	private ClonableProperty(ExtendedProperty p)
	{
		super(p.getName(), p.getTitle(), p.getDescription(), p.getColumn(), 
			p.getType(), p.getLength(), p.isNullable(), true, true, 
			false, p.getAverageMethod(), null);
		this.isExtendedProperty = true;
		this.isMandatory = false;
	}
	
	/**
		Create a cloned property representation of a data-layer 
		object.
		@param propertyName The name of the property
		@param d Other metadata for the property
	*/
	ClonableProperty(String propertyName, ReporterClonePropertyData d)
	{
		super(propertyName, d.getTitle(), d.getDescription(), d.getColumn(),
			Type.fromValue(d.getValueType()), d.getLength(), d.isNullable(), true, true,
			false, AverageMethod.fromValue(d.getAverageMethod()), null);
		this.isExtendedProperty = d.isExtendedProperty();
		this.isMandatory = d.isMandatory();
	}
	
	/**
		Create a data-layer reprsentation of the cloned property.
	*/
	ReporterClonePropertyData toData()
	{
		ReporterClonePropertyData data = new ReporterClonePropertyData();
		data.setTitle(getTitle());
		data.setDescription(getDescription());
		data.setColumn(getColumn());
		data.setValueType(getType().getValue());
		data.setLength(getLength());
		data.setNullable(isNullable());
		data.setAverageMethod(getAverageMethod().getValue());
		data.setExtendedProperty(isExtendedProperty());
		data.setMandatory(isMandatory());
		return data;
	}

	/**
		Is this cloned property an extended property or a "static" property?
	*/
	public boolean isExtendedProperty()
	{
		return isExtendedProperty;
	}
	
	/**
		Check if this property is mandatory in all clone templates.
	*/
	public boolean isMandatory()
	{
		return isMandatory;
	}
	
	/**
		Check if this cloned property still exists. A cloned property that
		has been added to a template is not automatically removed if the
		extended properties configuration is changed.
	*/
	public boolean exists()
	{
		return !isExtendedProperty() || ExtendedProperties.getProperty("ReporterData", getName()) != null;
	}
	
	/*
		From the Object class
		---------------------
	*/
	@Override
	public String toString()
	{
		return getName();
	}

	@Override
	public boolean equals(Object o)
	{
		if (this == o) return true;
		if (o == null) return false;
		if (o.getClass() != ClonableProperty.class) return false;
		ClonableProperty other = (ClonableProperty)o;
		return getName().equals(other.getName());
	}
	
	@Override
	public int hashCode()
	{
		return getName().hashCode();
	}
	// ---------------------
	/*
		From the Comparable interface
		-----------------------------
	*/
	/**
		Sort by title, but mandatory properties before non-madatory.
	*/
	@Override
	public int compareTo(ClonableProperty o)
	{
		if (isMandatory() != o.isMandatory())
		{
			return isMandatory() ? -1 : 1;
		}
		return getTitle().compareTo(o.getTitle());
	}
	// -----------------------------
}
