/*
	$Id: DateUtil.java 7308 2017-03-13 08:15:46Z nicklas $

	Copyright (C) 2005 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

/**
	Utility methods that will make it easier to handle date values.
	Especially, dates must be cloned whenver they are returned to
	or passed from client applications.

	@author Nicklas
	@version 2.0
*/
public class DateUtil
{

	/**
		Make a copy of a <code>Date</code> object. Null values are allowed.

		@param value The date to check
		@param name The name of the attribute to use if an exception is thrown
		@return A copy of the original date
	*/
	public static Date setNullableDate(Date value, String name)
	{
		return copy(value);
	}

	/**
		Make a copy of a <code>Date</code> object. Null values are not allowed.

		@param value The date to check. Can not be null.
		@param name The name of the attribute to use if an exception is thrown
		@return A copy of the original date
	 	@throws InvalidUseOfNullException If the value parameter is null.
	*/
	public static Date setNotNullDate(Date value, String name)
		throws InvalidUseOfNullException
	{
		if (value == null) throw new InvalidUseOfNullException(name);
		return copy(value);
	}

	/**
		Make a copy of the date. If null, today's date is
		used instead.
		@since 3.11
	*/
	public static Date setDateOrToday(Date value)
	{
		return value == null ? new Date() : copy(value);
	}

	/**
		Get a copy of the date. Null values are allowed.
	 	@param value Date value to copy 
	 	@return A cloned Date object or null if argument is null.
	*/
	public static Date copy(Date value)
	{
		return value == null ? null : (Date)value.clone();
	}
	
	/**
		Truncate the given date so that only the date part
		remains. The time part is set to 00:00:00.0
		@param value The date to truncate, can be null
		@return A new date with time part set to 0 or null
		@since 2.16
	*/
	public static Date truncate(Date value)
	{
		if (value == null) return null;
		Calendar cal = Calendar.getInstance();
		cal.setTime(value);
		cal.set(Calendar.HOUR, 0);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);
		return cal.getTime();
	}
	
	private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
	
	/**
	  	Parses a string to create a <code>Date</code> without time. This method supports
	  	date in yyyy-MM-dd format or as long timevalues.
	  	 
	  	@param value the <code>String</code> to be parsed
	  	@return a <code>Date</code> object
	  	@throws DateFormatException if <code>value</code> isn't a valid date.
	*/
	public static Date parseString(String value)
		throws InvalidDataException
	{
		Date result = null;
		if (value != null)
		{
			try
			{
				result = DATE_FORMAT.parse(value);
			}
			catch (Exception ex)
			{
				try
				{
					result = new Date(new Long(value));
				}
				catch (Exception ex2)
				{
					throw new DateFormatException("The value "+value+" is not a valid date for format: " + DATE_FORMAT.toPattern());
				}
			}
		}
		return result;
	}
	
	/**
		Formats a date in yyyy-MM-dd format.
	 	@param d The date to format. 
	 	@return Formated date as a String.
		@since 2.2
	*/
	public static String formatDate(Date d)
	{
		return DATE_FORMAT.format(d);
	}
	

	private static final SimpleDateFormat TIMESTAMP_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

	/**
	  	Parses a string to create a <code>Date</code> including time. This method 
	  	supports date in yyyy-MM-dd HH:mm:ss format or as long timevalues.
	  	 
	  	@param value the <code>String</code> to be parsed
	  	@return a <code>Date</code> object
	  	@throws DateFormatException if <code>value</code> isn't a valid timestamp
	  	@since 2.16
	*/
	public static Date parseTimestamp(String value)
		throws InvalidDataException
	{
		Date result = null;
		if (value != null)
		{
			try
			{
				result = TIMESTAMP_FORMAT.parse(value);
			}
			catch (Exception ex)
			{
				try
				{
					result = new Date(new Long(value));
				}
				catch (Exception ex2)
				{
					throw new DateFormatException("The value "+value+" is not a valid timestamp for format: " + TIMESTAMP_FORMAT.toPattern());
				}
			}
		}
		return result;
	}

	/**
		Formats a timestamp in yyyy-MM-dd HH:mm:ss format.
	 	@param d The date to format. 
	 	@return Formated date as a String
	 	@since 2.16
	*/
	public static String formatTimestamp(Date d)
	{
		return TIMESTAMP_FORMAT.format(d);
	}

	/**
		Compare two date instances for equality while protecting from
		strange implementations in the java.util.Date and java.sql.Timestamp
		classes.
		@param time TRUE to compare the date+time part, FALSE to only compare the
			date part
		@since 3.9.1
	*/
	public static boolean safeEquals(Date d1, Date d2, boolean time)
	{
		if (d1 == d2) return true;
		if (d1 == null || d2 == null) return false;
		if (!time)
		{
			d1 = truncate(d1);
			d2 = truncate(d2);
		}
		return d1.getTime() == d2.getTime();
	}
	
	/**
		Checks if two lists contains equal elements disregarding
		their positions in the array. Eg. [1, 2] is equal to [2, 1]
		@param time TRUE to compare the date+time part, FALSE to only compare the
			date part
		@return TRUE if both list contains the same elements
		@since 3.9.1
	*/
	public static boolean safeInvariantEquals(List<Date> l1, List<Date> l2, boolean time)
	{
		if (l1 == l2) return true;
		if (l1 == null || l2 == null) return false;
		if (l1.size() != l2.size()) return false;
		
		boolean[] matched = new boolean[l2.size()];
		for (Date d1 : l1)
		{
			boolean found = false;
			for (int j = 0; j < l2.size(); ++j)
			{
				if (!matched[j] && safeEquals(d1, l2.get(j), time))
				{
					matched[j] = true;
					found = true;
					break;
				}
			}
			if (!found) return false;
		}
		return true;
	}

	
}
