/*
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;

import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;
import net.sf.basedb.core.data.ItemListData;
import net.sf.basedb.core.data.ItemListSyncFilterData;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.util.listable.ListableUtil;
import net.sf.basedb.util.listable.SourceItemTransformer;
import net.sf.basedb.util.listable.SourceItemTransformerFactory;
import net.sf.basedb.util.listable.TransformContext;

/**
	This class represents a list of items. Supported items to put
	in the list are those that implement the {@link Listable} 
	interface.

	@author Nicklas
	@since 3.5
*/
public class ItemList
	extends AnnotatedItem
{
	/**
		The type of item represented by this class.
		@see Item#ITEMLIST
		@see #getType()
	*/
	public static final Item TYPE = Item.ITEMLIST;
	
	/**
		The maximum length of the external id variable that can be stored in the
		database. Check the length against this value before calling the
		{@link #setExternalId(String)} method to avoid exceptions.
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = ItemListData.MAX_EXTERNAL_ID_LENGTH;

	
	/**
		Create a new <code>ItemList</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param memberType The item type of members in this list
		@return The new <code>ItemList</code> item
		@throws BaseException If there is an error
	*/
	public static ItemList getNew(DbControl dc, Item memberType)
		throws BaseException
	{
		ItemList l = dc.newItem(ItemList.class);
		l.setName("New item list");
		l.setMemberType(memberType);
		return l;
	}

	/**
		Get an <code>ItemList</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>ItemList</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static ItemList getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ItemList l = dc.loadItem(ItemList.class, id);
		if (l == null) throw new ItemNotFoundException("ItemList[id="+id+"]");
		return l;
	}

	/**
		Get a query that returns item lists.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ItemList> getQuery()
	{
		return new ItemQuery<ItemList>(ItemList.class);
	}

	/**
		Create a set with the intersection of the items
		in the given lists.
		@return A set with IDs of member items. The set is always a new
			set, modifications to it will not affect anything else
	*/
	private static Set<Integer> intersection(Collection<ItemList> lists, DelayedAdditionManager delayed)
	{
		Set<Integer> intersection = new HashSet<Integer>();
		if (lists != null)
		{
			boolean first = true;
			for (ItemList list : lists)
			{
				Set<Integer> other = list.getData().getMembers();
				if (first)
				{
					intersection.addAll(other);
					if (list.delayed != null)
					{
						delayed.queue().addAll(list.delayed.queue());
					}
					first = false;
				}
				else
				{
					intersection.retainAll(other);
					if (list.delayed != null)
					{
						delayed.queue().retainAll(list.delayed.queue());
					}
				}
			}
		}
		return intersection;
	}
	
	/**
		Create a set with the union of the items 
		in the given lists.
		@param delayed A temporary delayed additions manager to be used for collecting the
			union of delayed additions
		@return A set with IDs of member items. The set is always a new
			set, modifications to it will not affect anything else
	*/
	private static Set<Integer> union(Collection<ItemList> lists, DelayedAdditionManager delayed)
	{
		Set<Integer> union = new HashSet<Integer>();
		if (lists != null)
		{
			for (ItemList list : lists)
			{
				union.addAll(list.getData().getMembers());
				if (list.delayed != null)
				{
					delayed.queue().addAll(list.delayed.queue());
				}
			}
		}
		return union;
	}
	
	/**
		Count the number of times each item is present in the
		given lists and return those items that are found in
		&gt;=minCount and &lt;=maxCount lists.
		<p>
		NOTE! 
		<ul>
			<li>'minCount=1' and 'maxCount=number of lists' is the same
				as {@link #union(Collection, DelayedAdditionManager)}
			<li>'minCount=maxCount=number of lists' is the same
				as {@link #intersection(Collection, DelayedAdditionManager)}.
		</ul>
	
		@param lists The lists to look in
		@param minCount The minimum number of list an item
			is required to be present in
		@param maxCount The maximum number of lists an item
			is allowed to the present in
		@param delayed A temporary delayed additions manager to be used for collecting the
			union of delayed additions
		@return A set with IDs of member items. It is always a new
			set, modifications to it will not affect anything else
	*/
	private static Set<Integer> count(Collection<ItemList> lists, final int minCount, final int maxCount, DelayedAdditionManager delayed)
	{
		Map<Integer, Integer> union = new HashMap<Integer, Integer>();
		Map<Listable, Integer> unionDelayed = new HashMap<Listable, Integer>();
		
		if (lists != null)
		{
			for (ItemList list : lists)
			{
				for (Integer id : list.getData().getMembers())
				{
					Integer count = union.get(id);
					count = count == null ? 1 : count+1;
					union.put(id, count);
				}
				
				if (list.delayed != null)
				{
					for (Listable l : list.delayed.queue())
					{
						Integer count = unionDelayed.get(l);
						count = count == null ? 1 : count+1;
						unionDelayed.put(l, count);
					}
				}
			}
		}
		
		Predicate countFilter = new Predicate()
			{
				@Override
				public boolean evaluate(Object o)
				{
					int count = (Integer)o;
					return minCount <= count && count <= maxCount;
				}
			};

		CollectionUtils.filter(union.values(), countFilter);
		CollectionUtils.filter(unionDelayed.values(), countFilter);
		delayed.queue().addAll(unionDelayed.keySet());
		
		return union.keySet();
	}
	
	ItemList(ItemListData listData)
	{
		super(listData);
	}
	
	@Override
	ItemListData getData()
	{
		return (ItemListData)getBasicData();
	}
	
	/**
	 	Always returns null
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents() 
		throws BaseException
	{
		return null;
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/**
		Get the external id for this list.
		@return A string with the external id of this item, or null
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	/**
		Set the external id for this list. Null is allowed, but
		the value must not be longer than the value specified by
		the {@link #MAX_EXTERNAL_ID_LENGTH} constant. The value doesn't have to be unique.
		@param externalId The new external id for this item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_EXTERNAL_ID_LENGTH}
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}

	/**
		Get the number of members in this list.
	*/
	public int getSize()
	{
		return getData().getSize();
	}

	/**
		Get the item type of members on this list.
		@return An Item object
	*/
	public Item getMemberType()
	{
		return Item.fromValue(getData().getMemberType());
	}
	
	private void setMemberType(Item memberType)
	{
		if (memberType == null) throw new InvalidUseOfNullException("memberType");
		if (!Listable.class.isAssignableFrom(memberType.getItemClass()))
		{
			throw new InvalidDataException("Unsupported member type: " + memberType.name());
		}
		getData().setMemberType(memberType.getValue());
	}
	
	/**
		Checks if the other type is the same type as the member type of
		this list.
	*/
	private void checkMemberType(Item otherType)
	{
		if (getMemberType() != otherType)
		{
			throw new InvalidDataException("List '" + getName() + "' can't contain " + otherType);
		}
	}
	
	/**
		Checks that the lists have the same member type as this list.
	*/
	private void checkMemberType(Collection<ItemList> lists)
	{
		Item memberType = getMemberType();
		for (ItemList list : lists)
		{
			if (memberType != list.getMemberType())
			{
				throw new InvalidDataException("List '" + getName() + "' can't contain " + list.getMemberType());
			}
		}
	}
	
	/**
		Get the subtype of items that can be members in this list. This is
		a recommendation only and is not enforced.
		@return A subtype or null if not important
	*/
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	
	/**
		Set the subtype of items that can be members in this list. The
		subtype must have the same member type as this list.
		@param subtype A subtype or null if the subtype is not important
	*/
	public void setItemSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype == null)
		{
			getData().setItemSubtype(null);
		}
		else
		{
			checkMemberType(subtype.getMainItemType());
			subtype.checkPermission(Permission.USE);
			getData().setItemSubtype(subtype.getData());
		}
	}
	
	/**
	 	Get the raw data type of raw bioassays that can be members in this list. 
		This is a recommendation only and is not enforced.
	  	@return A {@link RawDataType} object
	*/
	public RawDataType getRawDataType()
	{
		return RawDataTypes.getRawDataType(getData().getRawDataType());		
	}
	
	/**
	 	Set the raw data type of raw bioassays that can be members in this list. 
		This is a recommendation only and is not enforced. The member
		type must be set to {@link Item#RAWBIOASSAY}
	  	
	  	@param rawDataType the {@link RawDataType} to use as default.
	*/
	public void setRawDataType(RawDataType rawDataType)
	{
		checkPermission(Permission.WRITE);
		checkMemberType(Item.RAWBIOASSAY);
		getData().setRawDataType(rawDataType == null ? null : rawDataType.getId());
	}
	
	/**
		If this flag is set synchronization of members with sync filters is disabled.
		It is always possible to add/remove or modify sync filters but the
		{@link #resyncFilter(DbControl, SynchronizeOption, boolean, ProgressReporter)}
		method will throw an exception if this is set.
	*/
	public boolean getDisableSyncFilters()
	{
		return getData().getDisableSyncFilters();
	}
	
	/**
		Disable or enable synchronization of members with sync filters.
	*/
	public void setDisableSyncFilters(boolean disable)
	{
		checkPermission(Permission.WRITE);
		getData().setDisableSyncFilters(disable);
	}
	
	private void checkSyncFiltersAllowed()
	{
		if (getDisableSyncFilters())
		{
			throw new PermissionDeniedException("Sync filters has been disabled: " + this);
		}
	}

	
	/**
		If this flag is set manual modification of members is disabled.
	*/
	public boolean getDisableManualMembers()
	{
		return getData().getDisableManualMembers();
	}
	
	/**
		Disable or enable manual modification of members.
	*/
	public void setDisableManualMembers(boolean disable)
	{
		checkPermission(Permission.WRITE);
		getData().setDisableManualMembers(disable);
	}

	private void checkManualMembersAllowed()
	{
		if (getDisableManualMembers())
		{
			throw new PermissionDeniedException("Manual modification of list members is disabled: " + this);
		}
	}
	
	private DelayedAdditionManager delayed;  

	private DelayedAdditionManager createDelayedAdditionManager()
	{
		DelayedAdditionManager mgr = new DelayedAdditionManager(getData());
		getDbControl().addTransactionalAction(mgr);
		return mgr;
	}
	
	private boolean delayedAdd(Listable item)
	{
		if (delayed == null) delayed = createDelayedAdditionManager();
		return delayed.add(item);
	}
	
	private boolean delayedRemove(Listable item)
	{
		return delayed == null || delayed.remove(item);
	}
	
	/**
		Add an item to this list.
	 
		@param item The item to add to the list
		@return TRUE if the item was added, FALSE if it already existed
			in this list
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the item is null or not of the
			correct member type
	*/
	public boolean add(Listable item)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (item == null) throw new InvalidUseOfNullException("item");
		checkMemberType(item.getType());
		boolean added = item.getId() == 0 ? delayedAdd(item) : getData().getMembers().add(item.getId());
		if (added) getData().setSize(getSize() + 1);
		return added;
	}
	
	/**
		Add multiple items to this list in one go.
		@param it An iterator that returns all items to be added,
			null items are ignored
		@return The numer of added items
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws InvalidDataException If the iterator returns an item
			of a different type than {@link #getMemberType()}
	*/
	public int add(Iterator<? extends Listable> it)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (it == null) return 0;
		int added = 0;
		Set<Integer> members = getData().getMembers();
		while (it.hasNext())
		{
			Listable item = it.next();
						
			if (item == null)
			{
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}
			
			checkMemberType(item.getType());
			if (item.getId() == 0)
			{
				if (delayedAdd(item)) added++;
			}
			else
			{
				if (members.add(item.getId())) added++;
			}
		}
		getData().setSize(getSize() + added);
		return added;
	}
	
	/**
		Remove an item from the list.
		@param item The item to remove.
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the item is null
	*/
	public void remove(Listable item)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (item == null) throw new InvalidUseOfNullException("item");
		checkMemberType(item.getType());
		boolean removed = item.getId() == 0 ? delayedRemove(item) : getData().getMembers().remove(item.getId());
		if (removed) getData().setSize(getData().getSize()-1);
	}
	
	/**
		Get a query that returns all items that are members of this list.
		@return An {@link ItemQuery} object
		@throws BaseException If the query could not be created and configured.
	*/
	public ItemQuery<? extends Listable> getMembers()
		throws BaseException
	{
		ItemQuery<? extends Listable> query = getAllItems();
		query.restrictPermanent(
			Hql.restriction("$id=ANY(SELECT mmb FROM ItemListData lst INNER JOIN lst.members mmb WHERE lst.id=" + getId()+")", "$")
		);
		return query;
	}
	
	/**
		Get a query that returns all items that can possibly be 
		members of this list. NOTE! This is just a convenience function
		that calls {@link BioSource#getQuery()}, etc. depending on the
		member type of the list.
		@return An {@link ItemQuery} object
		@throws BaseException If the query could not be created and configured
		@see Item#getQuery()
	*/
	@SuppressWarnings("unchecked")
	public ItemQuery<? extends Listable> getAllItems()
		throws BaseException
	{
		return (ItemQuery<? extends Listable>)getMemberType().getQuery();
	}

	/**
		Add members to this list.
		@param items The items to add
		@param other Delayed items to add
		@return The number of new items
	*/
	private int addMembers(Set<Integer> items, DelayedAdditionManager other)
	{
		ItemListData data = getData();
		Set<Integer> members = data.getMembers();
		int size = members.size();
		members.addAll(items);
		int numAdded = members.size() - size;
		
		if (other != null)
		{
			if (delayed == null) delayed = createDelayedAdditionManager();
			numAdded += delayed.addAll(other);
		}
		
		data.setSize(getSize() + numAdded);
		return numAdded;
	}
	
	/**
		Remove members from this list
		@param items The items to remove
		@return The number of removed items
	*/
	private int removeMembers(Set<Integer> items, DelayedAdditionManager other)
	{
		ItemListData data = getData();
		Set<Integer> members = data.getMembers();
		int size = members.size();
		members.removeAll(items);
		int numRemoved = size - members.size();
		
		if (other != null && delayed != null)
		{
			numRemoved += delayed.removeAll(other);
		}
		
		data.setSize(getSize() - numRemoved);
		return numRemoved;
	}
	
	/**
		Keep only specified members in this list
		@param items The items to keep
		@return The number of removed items
	*/
	private int retainMembers(Set<Integer> items, DelayedAdditionManager other)
	{
		int numRemoved = 0;
		ItemListData data = getData();
		Set<Integer> members = data.getMembers();
		int size = members.size();
		if (items == null || items.size() == 0)
		{
			members.clear();
		}
		else
		{
			members.retainAll(items);
		}
		numRemoved = size - members.size();
		
		if (delayed != null)
		{
			numRemoved = delayed.retainAll(other);
		}
		
		data.setSize(getSize() - numRemoved);
		return numRemoved;
	}
	
	/**
		Add all items from another item list to this
		list. The new list is the union of the existing list
		and the added list. The other list must have the 
		same member type as this list.
		
		@param list The list to add items from
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addAll(ItemList list)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (list == null) return 0;
		checkMemberType(list.getMemberType());
		return addMembers(list.getData().getMembers(), list.delayed);
	}

	/**
		Add items that are present in at least ONE OF the given lists to
		this list. Eg. the added items is the union from 
		the given lists. 
	
		@param lists The lists to add items from
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addUnion(Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (lists == null) return 0;
		checkMemberType(lists);

		int numAdded = 0;
		if (lists.size() == 1)
		{
			ItemList list = lists.iterator().next();
			numAdded = addMembers(list.getData().getMembers(), list.delayed);
		}
		else if (lists.size() > 1)
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numAdded = addMembers(union(lists, tmp), tmp);
		}
		return numAdded;
	}

	/**
		Add items that are present in the specified number of lists to 
		this list.
		<ul>
		<li>Delegate to {@link #addUnion(Collection)} if
			<code>minCount &lt;= 1 &amp;&amp; maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #addIntersection(Collection)}
			if <code>minCount = size(lists) &amp;&amp; maxCount &gt;= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount &gt; maxCount</code> no 
			items will be added.
		</ul>
		
		@param lists The lists to add items from
		@param minCount The minimum number of lists an item is required to
			be present in
		@param maxCount The maximum number of lists an item is allowed to 
			be present in
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addIfPresentIn(int minCount, int maxCount, Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (lists == null) return 0;
		checkMemberType(lists);
		
		int numAdded = 0;
		int numLists = lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numAdded = addUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numAdded = addIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numAdded = addMembers(count(lists, minCount, maxCount, tmp), tmp);
		}
		return numAdded;
	}

	/**
		Add items that are present in ALL OF the given lists to
		this list. Eg. the added items is the intersection from 
		the given lists. 
	
		@param lists The lists to add items from
		@return The number of items added (not including 
			items that already was in this list)
	*/
	public int addIntersection(Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (lists == null) return 0;
		checkMemberType(lists);
		
		int numAdded = 0;
		if (lists.size() == 1)
		{
			ItemList list = lists.iterator().next();
			numAdded = addMembers(list.getData().getMembers(), list.delayed);
		}
		else if (lists.size() > 1)
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numAdded = addMembers(intersection(lists, tmp), tmp);
		}
		return numAdded;
	}

	/**
		Remove all items from this list that are present
		in the other list. 
		
		@param list The list that contains the items
			to remove
		@return The number of items removed from this list
	*/
	public int removeAll(ItemList list)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (list == null) return 0;
		return removeMembers(list.getData().getMembers(), list.delayed);
	}

	/**
		Remove items that are present in at least ONE OF the given lists from
		this list. Eg. the removed items is the union from 
		the given lists. 
	
		@param lists The lists that contains the items to remove
		@return The number of items removed from this list
	*/
	public int removeUnion(Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (lists == null) return 0;
		
		int numRemoved = 0;
		if (lists.size() == 1)
		{
			ItemList list = lists.iterator().next();
			numRemoved = removeMembers(list.getData().getMembers(), list.delayed);
		}
		else if (lists.size() > 1)
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numRemoved = removeMembers(union(lists, tmp), tmp);
		}
		return numRemoved;
	}

	/**
		Remove items that are present in the specified number of lists from 
		this list.
		<ul>
		<li>Delegate to {@link #removeUnion(Collection)} if
			<code>minCount &lt;= 1 &amp;&amp; maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #removeIntersection(Collection)}
			if <code>minCount = size(lists) &amp;&amp; maxCount &gt;= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount &gt; maxCount</code> no 
			items will be removed.
		</ul>
		
		@param lists The lists with the items
		@param minCount The minimum number of lists an item is required to
			be present in
		@param maxCount The maximum number of lists an item is allowed to 
			be present in
		@return The number of items removed
	*/
	public int removeIfPresentIn(int minCount, int maxCount, Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (lists == null) return 0;
		
		int numRemoved = 0;
		int numLists = lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numRemoved = removeUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numRemoved = removeIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numRemoved = removeMembers(count(lists, minCount, maxCount, tmp), tmp);
		}
		return numRemoved;
	}

	/**
		Remove items that are present in ALL OF the given lists from
		this list. Eg. the removed items is the intersection from 
		the given lists. 
	
		@param lists The lists that contains the items to remove
		@return The number of items removed from this list
	*/
	public int removeIntersection(Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		if (lists == null) return 0;
		
		int numRemoved = 0;
		if (lists.size() == 1)
		{
			ItemList list = lists.iterator().next();
			numRemoved = removeMembers(list.getData().getMembers(), list.delayed);
		}
		else if (lists.size() > 1)
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numRemoved = removeMembers(intersection(lists, tmp), tmp);
		}
		return numRemoved;
	}

	/**
		Remove all items from this list that are NOT present
		in the other list. The new list is the intersection
		of the existing list and the other list. NOTE! If the
		other list is null or empty, all items will be removed
		from this list.
		
		@param list The list that contains the items
			to be retained
		@return The number of items removed from this list
	*/
	public int retainAll(ItemList list)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		return retainMembers(list == null ? null : list.getData().getMembers(), list == null ? null : list.delayed);
	}

	/**
		Retain items that are present in at least ONE OF the given lists 
		in this list. Eg. items that are not found in any of the given
		lists are removed from this list.
	
		@param lists The lists that contains the items to retain
		@return The number of items removed from this list
	*/
	public int retainUnion(Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		int numRemoved = 0;
		if (lists == null || lists.size() == 0)
		{
			numRemoved = retainMembers(null, null);
		}
		else if (lists.size() == 1)
		{
			ItemList list = lists.iterator().next();
			numRemoved = retainMembers(list.getData().getMembers(), list.delayed);
		}
		else
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numRemoved = retainMembers(union(lists, tmp), tmp);
		}
		return numRemoved;
	}

	/**
		Retain items that are present in ALL OF the given lists 
		in this list. Eg. items that are not found in all 
		lists of the given lists are removed from this list.
	
		@param lists The lists that contains the items to retain
		@return The number of items removed from this list
	*/
	public int retainIntersection(Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		int numRemoved = 0;
		if (lists == null || lists.size() == 0)
		{
			numRemoved = retainMembers(null, null);
		}
		else if (lists.size() == 1)
		{
			ItemList list = lists.iterator().next();
			numRemoved = retainMembers(list.getData().getMembers(), list.delayed);
		}
		else
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numRemoved = retainMembers(intersection(lists, tmp), tmp);
		}
		return numRemoved;
	}

	/**
		Retain items that are present in the specified number of lists in 
		this list.
		<ul>
		<li>Delegate to {@link #retainUnion(Collection)} if
			<code>minCount &lt;= 1 &amp;&amp; maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #retainIntersection(Collection)}
			if <code>minCount = size(lists) &amp;&amp; maxCount &gt;= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount &gt; maxCount</code> no 
			items will be removed.
		</ul>
		
		@param lists The lists with the items
		@param minCount The minimum number of lists an item is required to
			be present in
		@param maxCount The maximum number of lists an item is allowed to 
			be present in
		@return The number of items removed
	*/
	public int retainIfPresentIn(int minCount, int maxCount, Collection<ItemList> lists)
	{
		checkPermission(Permission.WRITE);
		checkManualMembersAllowed();
		int numRemoved = 0;
		int numLists = lists == null ? 0 : lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numRemoved = retainUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numRemoved = retainIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			DelayedAdditionManager tmp = new DelayedAdditionManager(null);
			numRemoved = retainMembers(count(lists, minCount, maxCount, tmp), tmp);
		}
		return numRemoved;
	}
	
	/**
		Get the last sync date+time. Note that manual changes may have been made to
		the list after this date.
	*/
	public Date getSyncDate()
	{
		return DateUtil.copy(getData().getSyncDate());
	}
	
	/**
		Does this item list define any sync filters?
		@since 3.6
	*/
	public boolean hasSyncFilters()
	{
		return getData().getSyncFilters().size() > 0;
	}
	
	/**
		Get a query returning all synchronization filters
		for this list
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<SyncFilter> getSyncFilters()
	{
		return SyncFilter.getQuery(this);
	}
	
	/**
		Checks if all enabled sync filters appears to be in sync or not.
		Note that disabled filters are not checked.
		@since 3.6
	*/
	public boolean isInSync(DbControl dc)
	{
		for (SyncFilter sf : getSyncFilters().list(dc))
		{
			if (!sf.isDisabled() && !sf.isInSync(dc)) return false;
		}
		return true;
	}
	
	/**
		Creates a new sync filter for this list. The sync filter will
		start using the same item type as the list member type and 
		{@link SourceItemTransform#NONE}. Use {@link SyncFilter#setSyncContext(ItemContext, SourceItemTransform)}
		to apply different sync filter and options.
		
	*/
	public SyncFilter newSyncFilter(DbControl dc)
	{
		checkPermission(Permission.WRITE);
		
		ItemContext cc = dc.getSessionControl().getCurrentContext(getMemberType());
		
		SyncFilter sf = dc.newItem(SyncFilter.class);
		ItemListSyncFilterData sfData = sf.getData();
		sf.setName("New sync filter");
		sfData.setItemList(getData());
		sfData.setSourceItemTransform(SourceItemTransform.NONE.ordinal());
		sfData.setSourceMustExist(true);
		sfData.setMatchingSource(-1);
		sfData.setMatchingTransformed(-1);
		cc.copyToSyncFilter(sfData);
		
		getData().getSyncFilters().add(sfData);
		return sf;
	}
	
	/**
		Copy all sync filters and list members from the given list.
		This is only allowed if the list to copy from is having the
		same member type as the current list. Existing members and
		sync filters in this list are removed.
	*/
	public void copySyncFiltersAndMembers(DbControl dc, ItemList copyFrom)
	{
		checkPermission(Permission.WRITE);
		checkSyncFiltersAllowed();
		if (copyFrom == null) return;
		checkMemberType(copyFrom.getMemberType());

		getData().getSyncFilters().clear();
		for (ItemListSyncFilterData copy : copyFrom.getData().getSyncFilters())
		{
			SyncFilter sf = dc.newItem(SyncFilter.class);
			ItemListSyncFilterData sfData = sf.getData();
			
			sf.setName(copy.getName());
			sfData.setItemList(getData());
			sfData.setSourceItemTransform(copy.getSourceItemTransform());
			sfData.setSourceMustExist(copy.getSourceMustExist());
			sfData.setMatchingSource(copy.getMatchingSource());
			sfData.setMatchingTransformed(copy.getMatchingSource());
			sfData.setMd5(copy.getMd5());
			sfData.setSourceItemType(copy.getSourceItemType());
			sfData.setInclude(copy.getInclude());
			sfData.getPropertyFilters().putAll(copy.getPropertyFilters());
			
			getData().getSyncFilters().add(sfData);
			dc.saveItemIf(this, sf, false);
		}

		getData().getMembers().clear();
		getData().getMembers().addAll(copyFrom.getData().getMembers());
		getData().setSize(getData().getMembers().size());
		getData().setSyncDate(copyFrom.getSyncDate());
	}
	
	/**
		Get a query that return all item lists that are used
		by synchronization filters in this list. Synchronization
		filters that are currently not enabled are ignored.
		
		@since 3.6
	*/
	public ItemQuery<ItemList> getDependingLists()
	{
		Set<Integer> lists = new HashSet<Integer>();
		lists.add(0);// To protect against an empty parameter list
		
		for (ItemListSyncFilterData sf : getData().getSyncFilters())
		{
			if (sf.isDisabled()) continue;
			
			for (Map.Entry<String, PropertyFilterData> pf : sf.getPropertyFilters().entrySet())
			{
				if (pf.getKey().startsWith("§"))
				{
					lists.add(Integer.parseInt(pf.getValue().getValue()));
				}
			}
		}
		
		ItemQuery<ItemList> query = ItemList.getQuery();
		query.restrictPermanent(Restrictions.in(Hql.property("id"), Expressions.parameter("lists")));
		query.setParameter("lists", lists, Type.INT);
		query.setIncludes(Include.ALL);
		
		return query;
	}
	
	/**
		Re-sync with the current synchronization filter. If this list
		doesn't use a synchronization filter, no change is made.
		@return The number of members in the list after applying the filter
	*/
	public int resyncFilter(DbControl dc, SynchronizeOption syncOptions, boolean applySubtypeFilter, ProgressReporter progress)
	{
		checkPermission(Permission.WRITE);
		checkSyncFiltersAllowed();
		Set<ItemListSyncFilterData> syncFilters = getData().getSyncFilters();
		if (syncFilters == null || syncFilters.size() == 0) return getSize();
		
		if (progress != null) progress.display(0, "Initializing filters...");
		int numFilters = syncFilters.size();
		int filterNo = 0;
		int numBefore = getSize();
		
		Set<Integer> matchingSubtype = null;
		if (applySubtypeFilter)
		{
			numFilters++;
			filterNo++;
			ItemSubtype subtype = getItemSubtype();
			RawDataType rawDataType = getRawDataType();
			if (subtype != null || rawDataType != null)
			{
				ItemQuery<? extends Listable> query = getAllItems();
				query.setIncludes(Include.ALL);
				if (subtype != null)
				{
					if (progress != null) progress.display(5, "Processing filter: " + subtype.getName());
					query.restrict(Restrictions.eq(Hql.property("itemSubtype"), Hql.entity(subtype)));
				}
				else if (rawDataType != null)
				{
					if (progress != null) progress.display(5, "Processing filter: " + rawDataType.getName());
					query.restrict(Restrictions.eq(Hql.property("rawDataType"), Expressions.string(rawDataType.getId())));
				}
				matchingSubtype = new HashSet<Integer>(query.idList(dc));
				if (progress != null) progress.append("; " + matchingSubtype.size() + " " + getMemberType());
			}
		}
		
		// For parent/child transformations
		TransformContext tc = new TransformContext(dc);
		SourceItemTransformerFactory tfFactory = ListableUtil.getTransformerFactory(getMemberType());
		
		// This will eventually contain the final members matching all filters
		// we start out with items matching the subtype filter (if any)
		Set<Integer> matchingItems = matchingSubtype == null ? null : new HashSet<Integer>(matchingSubtype);
		// This will be filled with items matching a 'must-not-exist' filter and must be removed from the intersection
		Set<Integer> mustNotExist = new HashSet<Integer>();
		for (ItemListSyncFilterData sf : syncFilters)
		{
			filterNo++;
			ThreadSignalHandler.checkInterrupted();

			// Ignore disabled and empty filters
			if (sf.isDisabled() || sf.getPropertyFilters().isEmpty()) 
			{
				sf.setMatchingSource(-1);
				sf.setMatchingTransformed(-1);
				sf.setMd5(null);
				continue;
			}
			if (progress != null)
			{
				progress.display(5 + (90 * filterNo) / numFilters, "Processing filter: " + sf.getName());
			}
			
			Item sourceItemType = Item.fromValue(sf.getSourceItemType());
			
			ItemQuery<? extends BasicItem> query = sourceItemType.getQuery();
			ItemContext syncContext = new ItemContext(sourceItemType, sf);
			syncContext.configureQuery(dc, query, true);
			syncContext.getInclude();

			// Load all matching items and save their id:s
			SortedSet<Integer> sources = new TreeSet<Integer>(query.idList(dc));
			sf.setMatchingSource(sources.size());
			sf.setMd5(SyncFilter.toMd5(sources));
			
			// Handle parent/child transforms
			SourceItemTransform itemTransform = SourceItemTransform.values()[sf.getSourceItemTransform()];
			Set<Integer> matching = sources;
			if (itemTransform != SourceItemTransform.NONE)
			{
				if (progress != null) progress.append("; " + sources.size() + " " + sourceItemType);

				SourceItemTransformer tf = tfFactory == null ? 
					null : tfFactory.create(sourceItemType, itemTransform);
				if (tf == null)
				{
					throw new UnsupportedOperationException("SourceItemTransform: " + itemTransform.name() + " (" + sourceItemType.name() + " -> " + getMemberType().name() + ")");
				}
				tc.setInclude(syncContext.getInclude());
				matching = tf.transform(tc, sources);
				// Keep track of items that we never wish to match
				if (!sf.getSourceMustExist()) mustNotExist.addAll(matching);
			}
			// Not really neccessary, but the numbers we report makes more sense if we
			// exclude items not matching the expected subtype
			if (matchingSubtype != null) matching.retainAll(matchingSubtype); 
			sf.setMatchingTransformed(matching.size());

			if (progress != null) progress.append("; " + matching.size() + " " + getMemberType());
			
			if (itemTransform == SourceItemTransform.NONE || sf.getSourceMustExist())
			{
				// If this is the first filter we simple keep all matching items
				if (matchingItems == null)
				{
					matchingItems = matching;
				}
				else
				{
					// After that we only keep items matching the new filter
					matchingItems.retainAll(matching);
				}
			}
		}
		
		if (matchingItems == null && mustNotExist.size() > 0)
		{
			// If all we have is a not-exists filter we load all items as the matching items
			if (progress != null) progress.display(95, "Processing filter: not-exists");
			ItemQuery<? extends Listable> query = getAllItems();
			query.setIncludes(Include.ALL);
			matchingItems = new HashSet<Integer>(query.idList(dc));
			if (progress != null) progress.append("; " + matchingItems.size() + " " + getMemberType());
		}
		
		if (matchingItems != null)
		{
			// Remove items that have matched a 'must-not-exist' filter
			matchingItems.removeAll(mustNotExist);
			
			// Get a new set since it performs better when updating
			Set<Integer> members = new HashSet<Integer>(getData().getMembers());
			// Update the members set
			syncOptions.synchronize(members, matchingItems);
			getData().setMembers(members);
			getData().setSize(members.size());
		}
		getData().setSyncDate(new Date());
		return getSize();
	}	
	
	/**
		Options for synchronizing an existing item list with a current settings
		in an {@link ItemContext}. See {@link ItemList#resyncFilter(DbControl, SynchronizeOption, boolean, ProgressReporter)}.
	*/
	public enum SynchronizeOption
	{
		/**
			Synchronize the members so that it contains all of the items matching the filter
			and no items not matching the filter. This options can both remove and add members.
		*/
		FULL 
		{
			@Override
			void synchronize(Set<Integer> existing, Set<Integer> matching) 
			{
				existing.clear();
				existing.addAll(matching);
			}
		},
		
		/**
			Synchronize the members by adding new items matching the filter.
			This option will not remove existing members that doesn't match the filter.
		*/
		ADD_ONLY 
		{
			@Override
			void synchronize(Set<Integer> existing, Set<Integer> matching) 
			{
				existing.addAll(matching);
			}
		},
		
		/**
			Synchronize the members by removing items that doesn't match the filter.
			This option will not add items that matches the filter but are not found
			in the list.
		*/
		REMOVE_ONLY 
		{
			@Override
			void synchronize(Set<Integer> existing, Set<Integer> matching) 
			{
				existing.retainAll(matching);
			}
		};
		
		/**
			Synchronize the existing members with the items matching the
			current filter context. The synchonrization should modify
			the 'existing' set.
			@param existing A set containing the existing members of the item list
			@param matching A set containing the items matching the sync filter
		*/
		abstract void synchronize(Set<Integer> existing, Set<Integer> matching);
	}
	
	/**
		Handles items that should be added as members to the list, but
		since they were created in the same transaction doesn't have any
		ID until later in the commit phase.
		@since 3.5.1
	*/
	static class DelayedAdditionManager
		implements TransactionalAction
	{

		private final ItemListData list;
		private final Set<Listable> queue;

		DelayedAdditionManager(ItemListData list) 
		{
			this.list = list;
			this.queue = new HashSet<Listable>();
		}
		
		boolean add(Listable item)
		{
			return queue.add(item);
		}
		boolean remove(Listable item)
		{
			return queue.remove(item);
		}
		
		int addAll(DelayedAdditionManager other)
		{
			int size = queue.size();
			queue.addAll(other.queue);
			return queue.size() - size;
		}
		
		int removeAll(DelayedAdditionManager other)
		{
			int size = queue.size();
			queue.removeAll(other.queue);
			return size - queue.size();
		}
		
		int retainAll(DelayedAdditionManager other)
		{
			int size = queue.size();
			if (other == null)
			{
				queue.clear();
			}
			else
			{
				queue.retainAll(other.queue);
			}
			return size - queue.size();
		}
		
		Set<Listable> queue()
		{
			return queue;
		}
		
		@Override
		public void onBeforeCommit() 
		{
			// Process queue
			Set<Integer> members = list.getMembers();
			for (Listable item : queue)
			{
				if (item.getId() != 0)
				{
					members.add(item.getId());
				}
			}
			list.setSize(members.size());
			queue.clear();
		}
		
		@Override
		public void onAfterCommit() 
		{}

		@Override
		public void onRollback() 
		{}
		

	}
}
