/*
  $Id: PluginType.java 7308 2017-03-13 08:15:46Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;

import net.sf.basedb.core.data.PluginTypeData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.util.JarClassLoader;

/**

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-03-13 09:15:46 +0100 (må, 13 mar 2017) $
*/
public class PluginType
	extends BasicItem
	implements Nameable, Removable, Registered
{

	/**
		The type of item represented by this class.
		@see Item#PLUGINTYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.PLUGINTYPE;

	/**
		The maximum length of the interface name that can be stored in the
		database. Check the length against this value before calling the
		{@link #setInterfaceName(String)} method to avoid exceptions.
	*/
	public static final int MAX_INTERFACENAME_LENGTH = PluginTypeData.MAX_INTERFACENAME_LENGTH;
	
	/**
		The maximum length of the jar path name that can be stored in the
		database.
		@since 3.0
	*/
	public static final int MAX_JARFILE_LENGTH = PluginTypeData.MAX_JARFILE_LENGTH;

	/**
		Create a new <code>PluginType</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param interfaceName The name of an interface that a plugin
			must implement to be of this type
		@param jarFile Optional file name of the JAR file which contains the
			plugin class. The JAR file must be located in the plugins.dir 
			directory. If not specified the plugin must be in the classpath 
		@return The new <code>PluginDefinition</code> item
		@throws BaseException If there is an error
	*/
	public static PluginType getNew(DbControl dc, String interfaceName, String jarFile)
		throws BaseException
	{
		PluginType pt = dc.newItem(PluginType.class);
		pt.setName(interfaceName);
		pt.loadInterfaceInformation(jarFile, interfaceName);
		pt.getData().setEntryDate(new Date());
		return pt;
	}

	/**
		Get a <code>PluginType</code> object when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The id of the item to load.
		@return The <code>PluginType</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static PluginType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PluginType pt = dc.loadItem(PluginType.class, id);
		if (pt == null) throw new ItemNotFoundException("PluginType[id="+id+"]");
		return pt;
	}
	
	/**
		Get a <code>PluginType</code> object when you know the interface
		name it represents.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param interfaceName The interface name of the plugin type
		@return The <code>PluginType</code> item
		@throws ItemNotFoundException If an item with the specified 
			interface name is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static PluginType getByInterfaceName(DbControl dc, String interfaceName)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (interfaceName != null) interfaceName = interfaceName.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(
			dc.getHibernateSession(), "GET_PLUGINTYPE_FOR_INTERFACENAME");
		query.setParameter("interfaceName", interfaceName);
		PluginTypeData pt = HibernateUtil.loadData(PluginTypeData.class, query);
		if (pt == null) throw new ItemNotFoundException("PluginType[interfaceName="+interfaceName+"]");
		return dc.getItem(PluginType.class, pt);
	}
	
	/**
		Get a query configured to retrieve plugin types.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<PluginType> getQuery()
	{
		return new ItemQuery<PluginType>(PluginType.class);
	}
	
	/**
		Creates a plugin type item.
		
		@param data the data
	*/
	PluginType(PluginTypeData data)
	{
		super(data);
	}

	@Override
	PluginTypeData getData()
	{
		return (PluginTypeData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}

	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}

	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}

	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		All users get READ access.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		granted |= Permission.grant(Permission.READ);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().getRemovedBy() != null;
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemovedBy(removed ? getSessionControl().getLoggedInUserId() : null);
	}
	@Override
	public User getRemovedBy()
		throws PermissionDeniedException, ItemNotFoundException 
	{
		Integer userId = getData().getRemovedBy();
		return userId == null ? null : User.getById(getDbControl(), userId);
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/**
		Get the class name of the Java class that implements this plugin. The
		specified class must implement the {@link net.sf.basedb.core.plugin.Plugin}
		interface.
		@return The class name for this plugin
	*/
	public String getInterfaceName()
	{
		return getData().getInterfaceName();
	}
	/**
		The class name cannot be changed. Used internally only.
	*/
	private void setInterfaceName(String interfaceName)
		throws InvalidDataException
	{
		getData().setInterfaceName(StringUtil.setNotNullString(interfaceName, "interfaceName", MAX_INTERFACENAME_LENGTH));
	}

	/**
		Get the name of the JAR file that contains this plug-in.
		If this value is null, the plug-in must be on the classpath
		(eg. in WEB-INF/lib).
		@since 3.0
	*/
	public String getJarFile()
	{
		return getData().getJarFile();
	}

	/**
		Get the path to the JAR file that contains the plugin class file.
		If this value is null, the plugin must be on the classpath.
	*/
	public String getJarPath()
	{
		String file = getJarFile();
		if (file == null) return null;
		return new java.io.File(Application.getPluginsDirectory(), file).getAbsolutePath();
	}


	/**
		The jar path cannot be changed. Used internally only.
	*/
	private void setJarFile(String jarFile)
		throws InvalidDataException
	{
		getData().setJarFile(StringUtil.setNullableString(jarFile, "jarFile", MAX_JARFILE_LENGTH));
	}
	
	/**
	 	@param jarFile File name of the jar file containing the plug-in. The
	 		JAR file must be located in the directory spercified by 'plugins.dir' setting
	 		in base.config. Null is allowed if plugin is located in the classpath (eg. WEB-INF/lib).
	 	@param interfaceName The name of an interface that a plugin
			must implement to be of this type. 
	 	@throws InvalidDataException If the found class not really is an interface, or 
	 		if the found interface doesn't extend from {@link net.sf.basedb.core.plugin.Plugin}
	 	@throws BaseException If there is some other error.

	*/
	public void loadInterfaceInformation(String jarFile, String interfaceName)
		throws InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (jarFile != null)
		{
			if (!Path.isValidName(jarFile))
			{
				throw new InvalidDataException("JAR file name is not a valid name: " + jarFile);
			}
		}
		
		String jarPath = jarFile == null ? 
				null : new java.io.File(Application.getPluginsDirectory(), jarFile).getAbsolutePath();
		try
		{
			Class<?> c = null;
			if (jarPath == null)
			{
				c = Class.forName(interfaceName);
			}
			else
			{
				ClassLoader cl = JarClassLoader.getInstance(jarPath, true);
				c = cl.loadClass(interfaceName);
			}
			if (!c.isInterface())
			{
				throw new InvalidDataException("Class is not an interface: "+interfaceName);
			}
			if (!Plugin.class.isAssignableFrom(c))
			{
				throw new InvalidDataException("Interface does not extend net.sf.basedb.core.plugin.Plugin: "+interfaceName);
			}
			setInterfaceName(interfaceName);
			setJarFile(jarFile);
		}
		catch (Exception ex)
		{
			throw new BaseException("Could not instantiate plugin interface: "+interfaceName+"", ex);
		}
	}
	
	/**
		Get a query that returns the plugin definitions which implements
		this type.
		@see PluginDefinition#getQuery()
	*/
	public ItemQuery<PluginDefinition> getPlugins()
	{
		ItemQuery<PluginDefinition> query = PluginDefinition.getQuery();
		query.joinPermanent(Hql.innerJoin("pluginTypes", Item.PLUGINTYPE.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.PLUGINTYPE.getAlias()),
				Hql.entity(this)
			)
		);
		return query;
	}
}
