/*
	$Id: RemovableData.java 7015 2015-11-17 12:57:57Z nicklas $

	Copyright (C) 2005, 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

/**
	A removable item is an item that can be flagged
	as removed. This doesn't remove the information about the
	item from the database, but can be used by client applications
	to hide items that the user is not interested in.
	<p>
	For example, all methods generating lists of items, will by default
	not include items that are flagged as removed.
	<p>
	A cleaning application is usually run at regular intervals
	to remove items that have been flagged as removed.
	
	<p>
	This interface defines Hibernate database mapping for the
	<code>removedBy</code> property to the database column <code>removed_by</code>.
	The column is an integer column holding the ID of the user that
	flagged the item or null if the item has not been flagged.
	<p>
	If a subclass wants to map these properties to other columns, 
	it should override the {@link #getRemovedBy()} method and add a 
	Hibernate tag in the comment.
	<p>
	
	<b>Reference implementation</b><br>
	<pre class="code">
private Integer removedBy;
public Integer getRemovedBy()
{
   return removedBy;
}
public void setRemovedBy(Integer userId)
{
   this.removedBy = userId;
}
</pre>

	@author Nicklas
	@version 2.0
	@see <a href="../../../../../../html/developer/api/data_api.basic.html">Developer documentation: Basic classes and interfaces</a>
*/
public interface RemovableData
	extends IdentifiableData
{

	/**
		Get the ID of the user that removed this item.
		@return The ID of a user or null if the item is not removed
		@since 3.7
		@hibernate.property column="`removed_by`" type="int" not-null="false"
	*/
	public Integer getRemovedBy();

	/**
		Set the ID of the user that removed this item or null
		to restore the item.
		@since 3.7
	*/
	public void setRemovedBy(Integer userId);
	
}
