/**
	$Id: SchemaGenerator.java 7150 2016-04-27 06:38:21Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.util.EqualsHelper;

import org.hibernate.boot.Metadata;
import org.hibernate.boot.model.relational.Namespace;
import org.hibernate.boot.spi.MetadataImplementor;
import org.hibernate.dialect.Dialect;
import org.hibernate.engine.jdbc.connections.spi.JdbcConnectionAccess;
import org.hibernate.engine.jdbc.env.spi.JdbcEnvironment;
import org.hibernate.engine.jdbc.spi.JdbcServices;
import org.hibernate.jdbc.Work;
import org.hibernate.service.ServiceRegistry;
import org.hibernate.tool.schema.extract.internal.DatabaseInformationImpl;
import org.hibernate.tool.schema.extract.spi.DatabaseInformation;
import org.hibernate.tool.schema.internal.SchemaCreatorImpl;
import org.hibernate.tool.schema.internal.SchemaMigratorImpl;
import org.hibernate.tool.schema.spi.SchemaCreator;
import org.hibernate.tool.schema.spi.SchemaMigrator;
import org.hibernate.tool.schema.spi.Target;;

/**
	Replaces the Hibernate SchemaExport and SchemaUpdate tools. We need
	this to be able to intercept and modify some of the SQL statments that
	Hibernate generates since they don't always work as expected.
	
	* Since Hibernate 5 we only trust table and column definitions. Foreign keys
	and indexes are created with {@link DbIndexWork}.
	
	* Hibernate generate "not null" constraints for
	columns that are added to existing tables and we don't want that.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2016-04-27 08:38:21 +0200 (on, 27 apr 2016) $
*/
public class SchemaGenerator
	implements Work
{
	
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(SchemaGenerator.class);
	
	private final Metadata metadata;
	private final Dialect dialect;
	private final DbEngine dbEngine;
	private final Mode mode;
	private final ProgressReporter progress;
	private List<String> ignoredSql;
	
	/**
		Create a new schema generator. 
		
		@param metadata Information about current configuration
		@param dialect The dialect of the database we are connecting to
		@param dbEngine The DbEngine for the database we are connecting to
		@param mode The installation mode
		@param progress An optional progress reporter
	*/
	public SchemaGenerator(Metadata metadata, Dialect dialect, DbEngine dbEngine, Mode mode, ProgressReporter progress)
	{
		this.metadata = metadata;
		this.dialect = dialect;
		this.dbEngine = dbEngine;
		this.mode = mode;
		this.progress = progress;
	}

	/*
		From the Work interface
		-----------------------
	*/
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		log.info("Starting schema generator");
		log.info("Dialect: " + dialect);
		log.info("DbEngine: " + dbEngine);
		log.info("Mode: " + mode);
		
		this.ignoredSql = new ArrayList<String>();
		
		Statement stmt = connection.createStatement();
		try
		{
			final CollectSqlTarget schemaUpdates = new CollectSqlTarget(dbEngine, dialect, mode);
			final CollectSqlTarget indexUpdates = new CollectSqlTarget();
			
			if (mode == Mode.UPDATE)
			{
				log.info("Fetching database metadata");
				ServiceRegistry serviceRegistry = ((MetadataImplementor)metadata).getMetadataBuildingOptions().getServiceRegistry();
				JdbcServices jdbcServices = serviceRegistry.getService( JdbcServices.class );
				JdbcEnvironment jdbcEnvironment = jdbcServices.getJdbcEnvironment();
				JdbcConnectionAccess jdbcConnectionAccess = jdbcServices.getBootstrapJdbcConnectionAccess();
				Namespace ns = metadata.getDatabase().getDefaultNamespace();
				DatabaseInformation existingDatabase = new DatabaseInformationImpl(serviceRegistry, jdbcEnvironment, jdbcConnectionAccess, ns.getPhysicalName().getCatalog(), ns.getPhysicalName().getSchema());
				
				log.info("Generating schema update script");
				SchemaMigrator schemaMigrator = new SchemaMigratorImpl();
				schemaMigrator.doMigration(metadata, existingDatabase, false, Arrays.asList((Target)schemaUpdates));
			}
			else if (mode == Mode.INSTALL)
			{
				log.info("Generating schema creation script");
				SchemaCreator schemaCreator = new SchemaCreatorImpl();
				schemaCreator.doCreation(metadata, false, schemaUpdates);
			}
			else
			{
				throw new UnsupportedOperationException(mode.name());
			}
			
			// Collect SQL statements that are required for creating foreign keys and indexes
			DbIndexWork dbIndex = new DbIndexWork(metadata, dialect, dbEngine, false, true, false, false);
			dbIndex.setTarget(indexUpdates);
			dbIndex.execute(connection);
			
			List<SqlStatement> allSql = new ArrayList<SqlStatement>();
			allSql.addAll(schemaUpdates.getSql());// May contain null statements
			allSql.addAll(indexUpdates.getSql());

			int numStatements = 
				schemaUpdates.getNumStatements()+
				indexUpdates.getNumStatements(); // Only counts non-null statements
			int numDone = 0;
			int interval = Math.max(1, numStatements / 30);
			
			for (SqlStatement sql : allSql)
			{
				if (sql.modified != null)
				{
					if (sql.isModified())
					{
						log.debug("Original SQL: " + sql.original);
					}
					log.debug("Executing: " + sql.modified);

					stmt.executeUpdate(sql.modified);

					if (progress != null)
					{
						if (numDone % interval == 0)
						{
							progress.display((100 * numDone) / numStatements, "--" + StringUtil.trimString(sql.modified, 55));
						}
						else
						{
							progress.append(".");
						}
					}
					numDone++;
				}
				else
				{
					ignoredSql.add(sql.original);
					log.debug("Ignoring: " + sql.original);
				}
			}
		}
		catch (SQLException | RuntimeException ex)
		{
			log.error("Schema update failed", ex);
			throw ex;
		}
		finally
		{
			if (stmt != null) 
			{
				try
				{
					stmt.close();
				}
				catch (SQLException ex)
				{}
			}
		}
		log.info("Schema update complete" );
		
	}

	/**
		Get a list with the SQL statements that was ignored in the last run.
		@return A list or null if the generator hasn't been executed
		@since 3.1
	*/
	public List<String> getIgnoredSql()
	{
		return ignoredSql;
	}
	
	/**
		Helper class used for collecting SQL statements generated by Hibernate
		schema creation/update tools.
		@since 3.8
	*/
	static class CollectSqlTarget
		implements Target
	{
		private final List<SqlStatement> allSql;
		private final Dialect dialect;
		private final DbEngine dbEngine;
		private final Mode mode;
		private int numStatements;
		
		CollectSqlTarget()
		{
			this(null, null, null);
		}
		
		CollectSqlTarget(DbEngine dbEngine, Dialect dialect, Mode mode)
		{
			this.dbEngine = dbEngine;
			this.dialect = dialect;
			this.mode = mode;
			this.allSql = new ArrayList<SqlStatement>();
			this.numStatements = 0;
		}
		
		public List<SqlStatement> getSql()
		{
			return allSql;
		}
		
		public int getNumStatements()
		{
			return numStatements;
		}
		
		@Override
		public void release() 
		{}
		
		@Override
		public void prepare() 
		{}
		
		@Override
		public boolean acceptsImportScriptActions() 
		{
			return false;
		}
		
		@Override
		public void accept(String sql) 
		{
			String modifiedSql = dbEngine == null ? sql : dbEngine.inspectSchemaGenerationSQL(sql, dialect, mode);
			allSql.add(new SqlStatement(sql, modifiedSql));
			if (modifiedSql != null) numStatements++;
		}
	}
	
	public static class SqlStatement
	{
		final String original;
		final String modified;
		
		public SqlStatement(String sql)
		{
			this(sql, sql);
		}
		
		public SqlStatement(String original, String modified)
		{
			this.original = original;
			this.modified = modified;
		}
		
		boolean isModified()
		{
			return !EqualsHelper.equals(original, modified);
		}
	}
	
	/**
		The installation mode.
		@since 3.1
	*/
	public static enum Mode
	{
		/**
			Creating a fresh installation.
		*/
		INSTALL(true, false),
		/**
			Updating an existing installation.
		*/
		UPDATE(false, true);
		
		private final boolean requireEmptyDatabase;
		private final boolean verifySchemaVersion;
		private Mode(boolean requireEmptyDatabase, boolean verifySchemaVersion)
		{
			this.requireEmptyDatabase = requireEmptyDatabase;
			this.verifySchemaVersion = verifySchemaVersion;
		}
		
		/**
			Does this installation mode require an empty database or not?
		*/
		public boolean requireEmptyDatabase()
		{
			return requireEmptyDatabase;
		}
		
		/**
			Does this installation mode require a check against the schema version
			of the intalled database or not?
		*/
		public boolean verifySchemaVersion()
		{
			return verifySchemaVersion;
		}
	}
	
}
