/**
	$Id: AbstractSignalReceiver.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.WeakHashMap;

import net.sf.basedb.core.Application;

/**
	Abstract base class that is useful when implementing signal receivers.
	This class helps with keeping track of signal handlers and with generating
	return values for the {@link #registerSignalHandler(SignalHandler)} method.
	This class will generate values in the form of {@link java.net.URI}:s.
	<code>signal://handlerId@receiverId/?supportedSignals</code>. 
	
	<ul>
	<li>The <i>receiverId</i> part is given by the parameter in the 
		{@link #init(String)} method and must be set by the implementing 
		subclass. 
		
	<li>The <i>handlerId</i> part is given by calling {@link #getLocalSignalHandlerId(SignalHandler)}.
		This method may be overridden by subclasses that needs a different method of
		local ID generation. The important thing is that the ID is unique among all
		registered signal handlers.
		
	<li>The <i>supportedSignals</i> part is created by joining all signal ID:s 
		from {@link SignalHandler#getSupportedSignals()} into a comma-separated string.
	</ul>

	The {@link AbstractSignalTransporter} contains methods for parsing this scheme
	and make the information available to subclasses as a {@link java.net.URI} and
	as higher-level objects.
	<p>
	Subclasses should override the {@link #getGlobalSignalId(SignalHandler)} if they 
	want to use a different ID generation scheme.

	@author nicklas
	@version 2.6
	@base.modified $Date: 2015-05-12 11:27:08 +0200 (ti, 12 maj 2015) $
*/
public abstract class AbstractSignalReceiver
	implements SignalReceiver
{

	/**
		Log signals processing.
	*/
	private static final org.slf4j.Logger logger = 
		org.slf4j.LoggerFactory.getLogger(AbstractSignalReceiver.class);
	

	/**
		Maps local signal handler ID to -&gt; signal handler itself.
	*/
	private Map<String, SignalHandler> handlers;
	private Map<SignalHandler, String> handlerIds;
	private String receiverId;
	private Thread notifyThread;
	
	/**
		Create a new signal receiver.
	 */
	protected AbstractSignalReceiver()
	{}

	/*
		From the SignalReceiver interface
		-------------------------------------------
	*/
	/**
		Initialise the signal receiver. 
		@param receiverId The ID of the signal receiver, needed if the default 
			ID generation should be used
	*/
	@Override
	public void init(String receiverId)
	{
		handlers = Collections.synchronizedMap(new HashMap<String, SignalHandler>());
		handlerIds = new WeakHashMap<SignalHandler, String>();
		this.receiverId = receiverId;
		logger.info("Initializing signal receiver: id=" + receiverId);
	}
	/**
		Close this signal receiver. If a subclass overrides this method it
		should call <code>super.close()</code> if it wants to
		use the default shutdown notification to signal handlers. This
		implementation will start a separate notification thread that first sends
		the {@link Signal#SHUTDOWN} signal to all registered handlers, and then 
		{@link Signal#ABORT} signal to all that are still alive after the shutdown signal. 
		The current thread will wait at most the given time. If the number of registered handlers
		goes down to 0 before the time has ended the current thread will be awakened
		so it can continue. Note that the notification thread will usually finish in 
		a short time, but it may take longer for all worker thread of each job to
		react to the signal.
		@param wait Number of milliseconds to wait for registered jobs to abort,
			use a negative value to disable notification, and 0 to disable waiting
	*/
	@Override
	public void close(int wait)
	{
		logger.info("Closing signal receiver: receiver id=" + receiverId);
		if (handlers == null) return;
		if (wait >= 0 && handlers.size() > 0)
		{
			Thread abortAll = new Thread(
				new Runnable()
				{
					@Override
					public void run()
					{
						logger.info("Sending SHUTDOWN/ABORT to all handlers");
						sendToAll(new ShutdownSignalSender());
					}
				}
			);
			abortAll.setDaemon(true);
			abortAll.start();
			if (wait > 0)
			{
				try
				{
					logger.info("Sleeping for " + wait + " milliseconds");
					notifyThread = Thread.currentThread();
					Thread.sleep(wait);
					notifyThread = null;
				}
				catch (InterruptedException ex)
				{
					// Good, all handlers got the signal before timeout
				}
			}
			logger.info("Remaining active handlers: " + handlers.size());
		}
		handlers.clear();
		handlers = null;
	}

	/**
		Register a signal handler with this receiver.
		@param handler The signal handler to register
		@return The global ID of the signal handler
	*/
	@Override
	public String registerSignalHandler(SignalHandler handler)
	{
		if (notifyThread != null) throw new SignalException("Signal receiver is shutting down.");
		String localId = getLocalSignalHandlerId(handler);
		String globalId = getGlobalSignalId(handler);
		if (handlers != null) 
		{
			synchronized (handlers)
			{
				if (handlers.containsKey(localId))
				{
					throw new SignalException("A handler with ID '" + localId + "' is already registered: "
						+ handlers.get(localId));
				}
				handlers.put(localId, handler);
			}
		}

		logger.info("Register signal handler: recevier id = " + receiverId + 
			"; global id=" + globalId + "; local id=" + localId);
		logger.debug("Current number of registered signal handlers: " + 
			(handlers == null ? "0/0" : handlers.size()+"/"+handlerIds.size()));
		return globalId;
	}

	/**
		Unregister a signal handler.
		@param handler The signal handler to unregister
	*/
	@Override
	public void unregisterSignalHandler(SignalHandler handler)
	{
		String localId = getLocalSignalHandlerId(handler);
		logger.info("Unregister signal handler: recevier id = " + receiverId + 
			"; local id=" + localId);
		if (handlers != null) 
		{
			synchronized (handlers)
			{
				handlers.remove(localId);
				handlerIds.remove(handler);
			}
			// If called from the close() method() this is waiting for
			if (notifyThread != null && handlers.size() == 0)
			{
				logger.info("Interrupting closedown thread");
				notifyThread.interrupt();
			}
		}
		logger.debug("Current number of registered signal handlers: " + 
			(handlers == null ? "0/0" : handlers.size()+"/"+handlerIds.size()));
	}
	
	@Override
	public void sendToAll(SignalSender sender)
	{
		if (handlers == null) return;
		logger.info("Using " + sender + " to " + (handlers.size()) + " handlers");
		// Copy the signal handlers to a temporary collection,
		// since sending a signal may cause a handler to get 
		// unregistered, which may fail or block due to synchronization
		Collection<SignalHandler> temp = null;
		synchronized (handlers)
		{
			temp = new ArrayList<SignalHandler>(handlers.values());
		}
		for (SignalHandler handler : temp)
		{
			sender.sendTo(handler);
		}
	}
	// -------------------------------------------

	/**
		Get the receiver ID that was passed to the {@link #init(String)}
		method.
	*/
	protected String getReceiverId()
	{
		return receiverId;
	}
	
	/**
		Generate a signal ID string. This string is returned by
		the {@link #registerSignalHandler(SignalHandler)} method and is used
		in {@link SignalTransporter#init(String)} method to initialise
		a transporter object so that it can send signals to the specified handler.
		See the class documentation for a description of the format of the
		generated string. The string is of the format:
		<p>
		<code>signal://handlerId@receiverId/?supportedSignals</code>
		<p>
		See class description for detailed information.
		
		@param handler The signal handler to generate the ID for
		@return The signal handler ID
	*/
	protected String getGlobalSignalId(SignalHandler handler)
	{
		StringBuilder sb = new StringBuilder();
		sb.append("signal://");
		sb.append(getLocalSignalHandlerId(handler)).append("@");
		sb.append(getReceiverId()).append("/?");
		Collection<Signal> signals = handler.getSupportedSignals();
		if (signals != null)
		{
			boolean first = true;
			for (Signal signal : signals)
			{
				if (!first) sb.append(",");
				sb.append(signal.getId());
				first = false;
			}
		}
		return sb.toString();
	}
	
	/**
		Get the local signal handler id of the given signal handler.
		This implementation generates a unique ID for the handler.
		@param handler The handler to get the id for
		@return The local handler id
	*/
	protected String getLocalSignalHandlerId(SignalHandler handler)
	{
		if (handlerIds == null) return String.valueOf(System.identityHashCode(handler));
		String id = handlerIds.get(handler);
		if (id == null)
		{
			synchronized (handlers) 
			{
				while (true)
				{
					id = Application.generateRandomId(4);
					if (!handlers.containsKey(id)) break;
				}
				handlerIds.put(handler, id);
			}
		}
		return id;
	}
	
	/**
		Get the signal handler with a given ID.
		@param localId The local signal handler ID as 
			returned by the {@link #getLocalSignalHandlerId(SignalHandler)}
			method
		@return The signal handler, or null if no handler is found
	*/
	protected SignalHandler getSignalHandler(String localId)
	{
		return handlers == null ? null : handlers.get(localId);
	}
	
	/**
		Process a signal message. If the message can't be understood or
		if no handler can be found this method does nothing. The signal will
		be delivered to the signal handler in the current thread.
		@param message The message to process, the format of the message
			must be compatible with the message that {@link 
			AbstractSignalTransporter#generateSignalMessage(Signal)} generates
	*/
	protected void processSignalMessage(String message)
	{
		logger.debug("Processing signal message: " + message);
		if (message == null) return;
		try
		{
			URI uri = new URI(message);
			String localId = uri.getUserInfo();
			Signal signal = Signal.getSignal(uri.getQuery());
			SignalHandler signalHandler = getSignalHandler(localId);
			if (signalHandler != null) 
			{
				signalHandler.handleSignal(signal);
			}
			else
			{
				logger.warn("No signal handler found for id: " + localId);
			}
		}
		catch (Exception ex)
		{
			// Ignore invalid messages
			logger.warn("Could not process signal message: " + message, ex);
		}
	}
	
	
}
