/**
	$Id: CelValidator.java 5623 2011-05-06 12:09:13Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.affymetrix;

import affymetrix.fusion.cdf.FusionCDFData;
import affymetrix.fusion.cel.FusionCELData;
import net.sf.basedb.core.File;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.StringUtil;

/**
	Helper class for working with Affymetrix CEL files.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-05-06 14:09:13 +0200 (fr, 06 maj 2011) $
*/
public class CelValidator
{

	/**
		Create a new instance.
	*/
	public CelValidator()
	{}

	private void setHeader(RawBioAssay rba, String name, String value)
	{
		if (name == null || value == null) return;
		value = StringUtil.trimString(value, RawBioAssay.MAX_HEADER_VALUE_LENGTH);
		rba.setHeader(name, value);
	}

	/**
		Load a CEL file using the Affymetric Fusion SDK. This method
		checks that the file exists, and read all headers.
		
		@param celFile The file to load
		@return A FusionCELData object representing the CEL file
		@throws ItemNotFoundException If the actual file is not on the server
		@throws InvalidDataException If the file is not a CEL file
	*/
	public FusionCELData loadCelFile(File celFile)
	{
		if (celFile == null) throw new InvalidUseOfNullException("celFile");
		if (!celFile.getLocation().isDownloadable())
		{
			throw new ItemNotFoundException("Can't download file data for file '" + 
					celFile.getName() + "'; location=" + celFile.getLocation());
		}
		FusionCELData cel = new FusionCELData();
		cel.setFileName(celFile.getName());
		cel.setInputStream(celFile.getDownloadStream(0));
		if (!cel.readHeader())
		{
			throw new InvalidDataException("Could not read CEL file '" +
					celFile.getPath()+"': " + cel.getError());
		}
		return cel;
	}

	/**
		Copy metadata from the CEL file to the raw bioassay.
		This will set the mumber of spots and headers from the
		file.
		
		@param cel The CEL file
		@param rba The raw bioassay
	*/
	public void copyMetadata(FusionCELData cel, RawBioAssay rba)
	{
		rba.setNumFileSpots(cel.getCells());
		
		setHeader(rba, "Algorithm", cel.getAlg());
		for (int i = 0; i < cel.getNumberAlgorithmParameters(); ++i)
		{
			String name = cel.getAlgorithmParameterTag(i);
			setHeader(rba, "Algorithm parameter: " + name, 
					cel.getAlgorithmParameter(name));
		}
		setHeader(rba, "Dat header", cel.getDatHeader());
		setHeader(rba, "Chip type", cel.getChipType());
	}
	
	/**
		Reset metadata on the given raw bioassay. This set the
		number of spots to 0 and remove all headers.
		@param rba The raw bioassay
	*/
	public void resetMetadata(RawBioAssay rba)
	{
		rba.setNumFileSpots(0);
		rba.removeHeaders();
	}
	
	/**
		Check if the loaded CEL and CDF files matches. The chip type is
		verified. Unfortunately the chip type is not stored inside the CDF
		file, but has to be given as an extra parameter. Typically, this
		should be the name of the file. This method also verify that the number 
		rows and columns in the two files matches.
		
		@param cel The CEL fila data
		@param cdf The CDF file data
		@param cdfChipType The CDF chip type
	*/
	public void validateCelAndCdf(FusionCELData cel, FusionCDFData cdf, String cdfChipType)
		throws InvalidDataException
	{
		String celChipType = cel.getChipType();
		if (!cdfChipType.startsWith(celChipType))
		{
			throw new InvalidDataException("CEL chip type (" + celChipType + 
				") doesn't match CDF chip type (" + cdfChipType + ")");
		}
		
		if (cel.getRows() != cdf.getHeader().getRows() || cel.getCols() != cdf.getHeader().getRows())
		{
			throw new InvalidDataException("CEL size (rows=" + cel.getRows() + ", cols=" + cel.getCols() +
				") doesn't match CDF size (rows=" + 
				cdf.getHeader().getRows() + ", cols=" + cdf.getHeader().getCols()+")");
		}		
	}

	
	
}
