/**
	$Id: ClassMapErrorHandler.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.error;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
	Error handler that allows the registration of one or more
	error handler for subclasses to <code>Throwable</code>.
	<p>
	Use the {@link #addErrorHandler(Class, ErrorHandler)} to register
	error handlers for specific exceptions. When the {@link #handleError(Throwable)}
	is invoked it checks the class of the throwable and looks for error handlers
	for that class. If it can't find an error handler it will check the superclass
	and so on. If no error handler is found the orginal error is re-thrown.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (må, 20 apr 2015) $
*/
public class ClassMapErrorHandler
	implements ErrorHandler
{

	/**
		Map of registered error handlers.
	*/
	private Map<Class<? extends Throwable>, List<ErrorHandler>> errorHandlers;
	private final ErrorHandler defaultErrorHandler;

	/**
		Create a new <code>ClassMapErrorHandler</code>.
		
		@param defaultErrorHandler The default error handler to use if no
			other error handler is found, if null the error is re-thrown
	*/
	public ClassMapErrorHandler(ErrorHandler defaultErrorHandler)
	{
		this.defaultErrorHandler = defaultErrorHandler;
		this.errorHandlers = new HashMap<Class<? extends Throwable>, List<ErrorHandler>>();
	}
	
	/**
		Add an error handler for throwable:s of a specific class. The
		error handler will also handle errors of all subclasses unless
		a specific error handler exists for that subclass.
		
		@param t The class of throwable:s the error handler should handle
		@param handler The error handler
	*/
	public void addErrorHandler(Class<? extends Throwable> t, ErrorHandler handler)
	{
		List<ErrorHandler> handlers = errorHandlers.get(t);
		if (handlers == null) 
		{
			handlers = new ArrayList<ErrorHandler>();
			errorHandlers.put(t, handlers);
		}
		handlers.add(0, handler);
	}

	/*
		From the ErrorHandler interface
		-------------------------------------------
	*/
	@Override
	public boolean handleError(Throwable t) 
		throws Throwable
	{
		Class<?> c = t.getClass();
		List<ErrorHandler> handlers = null;
		do
		{
			handlers = errorHandlers.get(c);
			if (handlers != null)
			{
				for (ErrorHandler handler : handlers)
				{
					if (handler.handleError(t)) return true;
				}
				handlers = null;
			}
			c = c.getSuperclass();
		} while (c != null && handlers == null);
		if (defaultErrorHandler != null && defaultErrorHandler.handleError(t)) return true;
		throw t;
	}
	// -------------------------------------------

}
