/**
	$Id: NodeValidator.java 5818 2011-10-21 11:15:41Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.overview.Failure;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.loader.NodeLoader;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	A node validator is responsible for checking the "validity" of
	a node. The rules that are checked is varying from node to node
	depending on what type of item it represents. Two common 
	"errors" are missing items and that a user is denied access to
	an item. Since this means that there is no item to validate
	there are special methods for those two cases.
	<p>
	
	The validation of a node is part of the node creation and happens
	in one of the {@link NodeFactory#createMissingNode(Node)},
	{@link NodeFactory#createDeniedNode(Node)} or {@link 
	NodeFactory#createNode(net.sf.basedb.core.BasicItem, Object, Node, ChildNodeDirection)}
	methods. The validation takes place in two steps. The first step calls
	one of the three <em>pre</em> methods:
	<ul>
	<li>{@link #preMissingItem(DbControl, OverviewContext, Node)}
	<li>{@link #preDeniedItem(DbControl, OverviewContext, Node)}
	<li>{@link #preValidate(DbControl, OverviewContext, Object, Node)}
	</ul>
	
	This happens <em>before</em> a node is created for an item. If the
	called method returns <code>false</code> no new node is created
	and further processing is stopped. The second step happens after
	the node has been created and calls one of the three <em>post</em>
	methods:
	<ul>
	<li>{@link #postMissingItem(DbControl, OverviewContext, Node, Node)}
	<li>{@link #postDeniedItem(DbControl, OverviewContext, Node, Node)}
	<li>{@link #postValidate(DbControl, OverviewContext, Node, Node)}
	</ul>

	<p>
	The last step is to validate a complete folder. This can't happen from
	the {@link NodeFactory} since it only deals with individual nodes. It
	is the responsibility of a {@link NodeLoader} to call {@link 
	#postValidateFolder(DbControl, OverviewContext, Node, Node)} after it has finished
	loading a folder node. NOTE! This should be called even if the folder node 
	is null, or if the folder doesn't contain any children (some validators may
	want to report on missing items!)

	<p>
	Validations failures should be registered with the current context
	using {@link OverviewContext#createFailure(Validator, Node, String, Fix...)}.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-21 13:15:41 +0200 (fr, 21 okt 2011) $
*/
public interface NodeValidator<I>
{

	/**
		Validate a missing item before creating a node for it. This usually means 
		creating a missing item {@link Failure} for the parent node and registering 
		it with the context.
		@param dc A DbControl to use if the validation needs to access the database
		@param context The current overview context
		@param parentNode The parent node that is missing the item
		@return TRUE if a node should be created, FALSE if not
	*/
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode);
	
	/**
		Validate an item that the current user doen't have access to before creating a 
		node for it. 
		@param dc A DbControl to use if the validation needs to access the database
		@param context The current overview context
		@param parentNode The parent node that is linked with the item
		@return TRUE if a node should be created, FALSE if not
	*/
	public boolean preDeniedItem(DbControl dc, OverviewContext context, Node parentNode);
	
	/**
		Validate an existing item before creating a node for it.
		@param dc A DbControl to use if the validation needs to access the database
		@param context The current overview context
		@param item The item to validate
		@param parentNode The parent node that is linked with the item
		@return TRUE if a node should be created, FALSE if not
	*/
	public boolean preValidate(DbControl dc, OverviewContext context, I item, Node parentNode);

	/**
		Validate a missing item after a node has been created for it.
		@param dc A DbControl to use if the validation needs to access the database
		@param context The current overview context
		@param node The node object that represents the missing item
		@param parentNode The parent node that is missing the item
	*/
	public void postMissingItem(DbControl dc, OverviewContext context, Node node, Node parentNode);

	/**
		Validate an item that the current user doen't have access to after a 
		node has been created for it. 
		@param dc A DbControl to use if the validation needs to access the database
		@param context The current overview context
		@param node The node object that represents the unaccessible item
		@param parentNode The parent node that is linked with the item
	*/
	public void postDeniedItem(DbControl dc, OverviewContext context, Node node, Node parentNode);
	
	/**
		Validate an existing item after a node has been created for it.
		@param dc A DbControl to use if the validation needs to access the database
		@param context The current overview context
		@param node The node object that represents item
		@param parentNode The parent node that is linked with the item
	*/
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode);

	/**
		Validate a folder node after all child (item-type) nodes has been
		added to the folder. This method must be called by node loaders
		in all cases were it is theoretically possible that a folder node
		can be created. Some validators may, for example, check on missing
		items!
		
		@param dc A DbControl to use for database access
		@param context The current overview context
		@param folderNode The folder node, or null if no folder has been created
		@param parentNode The parent node (never null)
	*/
	public void postValidateFolder(DbControl dc, OverviewContext context, Node folderNode, Node parentNode);

}
