/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Kit;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing kit items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@since 3.7
*/
public class KitImporter
	extends AbstractItemImporter<Kit>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.KIT, GuiContext.Type.LIST));
	
	/**
		Column mapping parameter for the kit expiration date.
	*/
	private static final PluginParameter<String> expirationDateColumnMapping = 
		new PluginParameter<String>(
			"expirationDateColumnMapping",
			"Expiration date",
			"Mapping that picks the expiration date from the data columns. The date format is " +
			"specified by the 'Date format' parameter." +
			"Example: \\Expiration date\\",
			optionalColumnMapping
		);

	/**
		Column mapping parameter for the kit inactive flag.
	*/
	private static final PluginParameter<String> inactiveColumnMapping = 
		new PluginParameter<String>(
			"inactiveColumnMapping",
			"Inactive",
			"Mapping that picks the inactive flag state from the data columns. "+
			"Example: \\Inactive\\",
			optionalColumnMapping
		);

	
	public KitImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper registeredMapper;
	private Mapper expirationDateMapper;
	private Mapper inactiveMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	@Override
	protected Item getItemForSubtypes()
	{
		return Item.KIT;
	}
	
	/**
		Adds column mappings for name and description.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(subtypeColumnMapping);
		parameters.add(registeredColumnMapping);
		parameters.add(expirationDateColumnMapping);
		parameters.add(inactiveColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	/**
		Calls {@link Kit#getQuery()}.
	*/
	@Override
	protected ItemQuery<Kit> createItemQuery()
	{
		return Kit.getQuery();
	}
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? BioSource.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? BioSource.MAX_DESCRIPTION_LENGTH : null, null);
		expirationDateMapper = getMapper(ffp, (String)job.getValue("expirationDateColumnMapping"), null, null);
		inactiveMapper = getMapper(ffp, (String)job.getValue("inactiveColumnMapping"), null, null);
		registeredMapper = getMapper(ffp, (String)job.getValue("registeredColumnMapping"), null, null);
	}
	@Override
	protected Kit createItem(DbControl dc, FlatFileParser.Data data)
	{
		Kit kit = Kit.getNew(dc);
		if (registeredMapper != null) 
		{
			kit.setEntryDate(parseDate(registeredMapper.getValue(data)));
		}
		updateItem(dc, kit, data);
		return kit;
	}
	@Override
	protected void updateItem(DbControl dc, Kit kit, FlatFileParser.Data data)
	{
		if (nameMapper != null) kit.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) kit.setDescription(descriptionMapper.getValue(data));
		if (expirationDateMapper != null) kit.setExpirationDate(parseDate(expirationDateMapper.getValue(data)));
		if (inactiveMapper != null) kit.setInactive(Values.getBoolean(inactiveMapper.getValue(data)));

		updateItemSubtype(dc, kit, data);
	}
	// ------------------------------------
}
