/* $Id$
	------------------------------------------------------------------
	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	JavaScript functions for the Joust2 tree outline.

	@author Nicklas
	@since 3.3
*/

var Joust2 = function()
{
	var joust = {};
	var internal = {};

	/**
		Initialize and draw the Joust tree on
		the given element. The tree definitions must
		be stored as a JSON object in 'data-joust-tree' 
		attribute.
	*/
	joust.draw = function(element)
	{
		element = Doc.element(element);
		
		var tree = Data.json(element, 'joust-tree');
		if (!tree) return;
		
		var last = tree.length-1;
		for (var itemNo = 0; itemNo < tree.length; itemNo++)
		{
			var menuItem = tree[itemNo];
			
			// Genereate <div> tag for each menuItem
			var menuElement = internal.toElement(element, menuItem, '', itemNo==0, itemNo==last);
			var appendElement = menuElement.parentNode == null;
			if (appendElement) element.appendChild(menuElement);
			
			// Generate child elements with padding
			if (menuItem.children)
			{
				var pad = internal.padHtml(menuItem);
				var childDiv = internal.addChildren(menuItem, pad);
				if (appendElement) element.appendChild(childDiv);
			}
		}
	}
	
	/**
		Add child items to a menu element. It is assumed that
		this method is called on a menu element that currently
		has no children.
	 */
	joust.addChildren = function(menuElement, children)
	{
		menuElement = Doc.element(menuElement);
		
		var menuItem = menuElement.item;
		menuItem.children = children;
		menuItem.isLazy = false;
		
		var pad = internal.padHtml(menuItem);
		var childContainer = internal.addChildren(menuItem, menuItem.pad+pad);
		menuElement.parentNode.insertBefore(childContainer, menuElement.nextSibling);
	}
	
	/**
		Toggle the open/closed status of the given menu
		element. Sends a 'joust-expand' event to the
		node unless noEvent option is set.
	*/
	joust.toggle = function(menuElement, noEvent)
	{
		menuElement = Doc.element(menuElement);
		var menuItem = menuElement.item;
		
		if (!menuItem.isOpen && menuItem.isLazy)
		{
			Events.sendCustomEvent(menuElement, 'joust-lazy-init', {});
			return;
		}
		
		menuItem.isOpen = menuItem.isOpen ? 0 : 1;
		var childContainer = Doc.element(menuItem.id+'-children');
		if (childContainer)
		{
			Doc.addOrRemoveClass(childContainer, 'closed', !menuItem.isOpen);
		}
		joust.updateIconsAndText(menuElement);
		
		if (!noEvent)
		{
			Events.sendCustomEvent(menuElement, 'joust-expand', {});
		}
	}
	
	/**
		Select the given menu element in the joust tree.
		Sends a 'joust-select' event to the selected node
		unless noEvent option is set.
	 */
	joust.select = function(menuElement, noEvent)
	{
		menuElement = Doc.element(menuElement);
		var menuItem = menuElement.item;
		var tree = menuItem.tree;

		// De-select the current selected item
		if (tree.selectedItem)
		{
			tree.selectedItem.isSelected = false;
			joust.updateIconsAndText(tree.selectedItem.id);
		}
		
		// Make the given item the selected item
		tree.selectedItem = menuItem;
		menuItem.isSelected = true;
		joust.updateIconsAndText(menuElement);
		
		// Ensure parent path is open
		var parentItem = menuItem.parent;
		while (parentItem != null)
		{
			if (!parentItem.isOpen)
			{
				parentItem.isOpen = 1;
				Doc.removeClass(parentItem.id+'-children', 'closed');
				joust.updateIconsAndText(parentItem.id);
			}
			parentItem = parentItem.parent;
		}
		
		// Make sure the selected menu entry is visible, otherwise try to scroll the main <div>
		if (menuElement.offsetTop > tree.scrollTop + tree.offsetHeight || menuElement.offsetTop < tree.scrollTop)
		{
			tree.scrollTop = menuElement.offsetTop - 40;
		}
		
		// Send 'joust-select' event
		if (!noEvent)
		{
			Events.sendCustomEvent(menuElement, 'joust-select', {});
		}
	}
	
	/**
		Re-draw icons based on the current state of the menu item
		represented by the element.
	*/
	joust.updateIconsAndText = function(menuElement)
	{
		menuElement = Doc.element(menuElement);
		var menuItem = menuElement.item;

		var outlineIcon = internal.outlineIcon(menuItem);
		if (outlineIcon)
		{
			Doc.element(menuItem.id+'-outline').src = outlineIcon.src;
		}
		
		var docIcon = internal.docIcon(menuItem);
		if (docIcon)
		{
			Doc.element(menuItem.id+'-icon').src = docIcon.src;
		}
		
		Doc.addOrRemoveClass(menuElement, 'selected', menuItem.isSelected);
	}
	
	/**
		Event handler for open/close of child items.
	 */
	internal.toggle = function(event)
	{
		event.cancelBubble = true;
		joust.toggle(event.currentTarget.parentNode);
	}
	
	/**
		Event handler for clicking on a menu item.
	*/
	internal.select = function(event)
	{
		event.cancelBubble = true;
		joust.select(event.currentTarget);
	}
	
	internal.addChildren = function(menuItem, pad)
	{
		// Create child <div> container
		var childElements = internal.makeDiv(menuItem.id+'-children', 'children ' + (menuItem.isOpen ? 'open' : 'closed'));
		if (menuItem.className) childElements.className += ' ' + menuItem.className;
	
		var tree = menuItem.tree;
		var children = menuItem.children;
		var last = children.length-1;
		
		for (var childNo = 0; childNo < children.length; childNo++)
		{
			// Create each child element
			var child = children[childNo];
			child.parent = menuItem;
			var childElement = internal.toElement(tree, child, pad, false, childNo==last);
			var appendElement = childElement.parentNode == null;
			if (appendElement) childElements.appendChild(childElement);
			
			// Recursive call to create children of the child
			if (child.children)
			{
				var childPad = internal.padHtml(child);
				var childDiv = internal.addChildren(child, pad+childPad);
				if (appendElement) childElements.appendChild(childDiv);
			}
		}
		return childElements;
	}
	
	/**
		Create a DOM element for the given menuItem definition.
	*/
	internal.toElement = function(tree, menuItem, pad, first, last)
	{
		// Remember some settings
		menuItem.pad = pad;
		menuItem.isFirst = first;
		menuItem.isLast = last;
		menuItem.tree = tree;
		menuItem.isSelected = false;
		
		// Create main menu <div> container
		var menuElement = internal.makeDiv(menuItem.id, 'joustitem', pad);
		menuElement.item = menuItem;
		if (menuItem.className) menuElement.className += ' ' + menuItem.className;
		menuElement.addEventListener('click', internal.select, false);
		
		// Outline icon
		var outlineIcon = internal.outlineIcon(menuItem);
		if (outlineIcon)
		{
			var outlineTag = outlineIcon.toHTML(menuItem.id+'-outline');
			var outline = internal.makeDiv(null, 'outline pad ' + (!last && !first ? 'line' : 'blank'), outlineTag);
			outline.addEventListener('click', internal.toggle, false);
			menuElement.appendChild(outline);
		}
		
		// Document icon
		var docIcon = internal.docIcon(menuItem);
		if (docIcon)
		{
			var iconTag = docIcon.toHTML(menuItem.id+'-icon');
			var icon = internal.makeDiv(null, 'icon', iconTag);
			menuElement.appendChild(icon);
		}
		
		// Menu text
		var text = internal.makeDiv(menuItem.id+'-text', 'text interactable', menuItem.text);
		if (menuItem.tooltip) text.title = menuItem.tooltip;
		menuElement.appendChild(text);
		
		return menuElement;
	}
	
	/**
		Get the outline icon that should be used for the menuItem
		in it's current state (eg. open/closed).
	 */
	internal.outlineIcon = function(menuItem)
	{
		if (menuItem.noOutlineIcon == true) return null;
		var name = '';
		if (menuItem.isLazy)
		{
			name += 'Lazy';
		}
		else if (menuItem.children && menuItem.children.length)
		{
			name += menuItem.isOpen ? 'Minus' : 'Plus';
		}
		else
		{
			name += 'Join';
		}
		
		if (menuItem.isLast && !menuItem.isFirst)
		{
			name += 'Bottom';
		}
		if (menuItem.isFirst && !menuItem.isLast)
		{
			name += 'Top';
		}
		return IconStore.getIcon(name);
	}
	
	internal.docIcon = function(menuItem)
	{
		return IconStore.getIconFromSet(menuItem.icon, menuItem.isOpen, menuItem.isSelected);
	}
	
	internal.padHtml = function(menuItem)
	{
		if (menuItem.noOutlineIcon == true) return '';
		return '<div class="pad '+(menuItem.isLast ? 'blank' : 'line')+'">&nbsp;</div>'
	}
	
	internal.makeDiv = function(id, className, innerHTML)
	{
		var div = null;
		if (id) div = Doc.element(id);
		if (!div) 
		{
			div = document.createElement('div');
			if (id) div.id = id;
			if (className) div.className = className;
			if (innerHTML) div.innerHTML = innerHTML;
		}
		else
		{
			if (className) div.className += ' ' + className;
			div.innerHTML = innerHTML ? innerHTML : '';
		}
		return div;
	}
	
	return joust;
}();



var IconStore = function()
{
	var iconStore = {};
	var iconSets = [];
	var icons = [];
	
	iconStore.addDefaultIcons = function()
	{
		// Icons for the outline
		iconStore.addIcon('Plus', new Icon('plus.png'));
		iconStore.addIcon('PlusTop', new Icon('plus-top.png'));
		iconStore.addIcon('PlusBottom', new Icon('plus-bottom.png'));
		
		iconStore.addIcon('Minus', new Icon('minus.png'));
		iconStore.addIcon('MinusTop', new Icon('minus-top.png'));
		iconStore.addIcon('MinusBottom', new Icon('minus-bottom.png'));

		iconStore.addIcon('Lazy', new Icon('lazy.png'));
		iconStore.addIcon('LazyTop', new Icon('lazy-top.png'));
		iconStore.addIcon('LazyBottom', new Icon('lazy-bottom.png'));

		iconStore.addIcon('Join', new Icon('join.png'));
		iconStore.addIcon('JoinTop', new Icon('join-top.png'));
		iconStore.addIcon('JoinBottom', new Icon('join-bottom.png'));

		iconStore.addIcon('Line', new Icon('line.png'));
		iconStore.addIcon('Blank', new Icon('blank.png'));
		
		// Icon sets for folder-like nodes
		iconStore.addIconSet('Folder', 'folder-closed.png', 'folder-selected.png', 'folder-open.png', 'folder-selected.png');
		iconStore.addIconSet('Root', 'root.png', 'root-selected.png');
		iconStore.addIconSet('Home', 'home.png', 'home-selected.png');
		iconStore.addIconSet('UserHome', 'userhome.png', 'userhome-selected.png');
		
		// Document nodes
		iconStore.addIconSet('Item', 'item.png', 'item-selected.png');
	}
	
	
	iconStore.addIconSet = function(name, closed, closedSelected, open, openSelected, path)
	{
		var set = {};
		set.closed = new Icon(closed, path);
		set.closedSelected = new Icon(closedSelected || closed, path);
		set.open = new Icon(open || closed, path);
		set.openSelected = new Icon(openSelected || closedSelected || open || closed, path);
		iconSets[name] = set;
		return set;
	}
	
	iconStore.getIconSet = function(name)
	{
		return iconSets[name];
	}
	
	iconStore.getIconFromSet = function(name, open, selected)
	{
		var set = iconSets[name];
		if (!set) return iconStore.nullIcon;
		
		var icon = null;
		if (open && selected)
		{
			icon = set.openSelected;
		}
		else if (open)
		{
			icon = set.open;
		}
		else if (selected)
		{
			icon = set.closedSelected;
		}
		else 
		{
			icon = set.closed;
		}
		return icon || iconStore.nullIcon;
	}
	
	iconStore.addIcon = function(name, icon)
	{
		icons[name] = icon;
	}
	
	iconStore.getIcon = function(name)
	{
		return icons[name] || iconStore.nullIcon;
	}
	
	iconStore.nullIcon = {};
	
	iconStore.nullIcon.toHTML = function()
	{
		return '';
	}
	
	return iconStore;
}();


function Icon(file, path)
{
	if (!path) path = App.getRoot() + 'images/joust/';
	this.src = path+file;

	// Preload the image
	this.img = new Image();
	this.img.src = this.src;
}

Icon.prototype.toHTML = function(id)
{
	var html = '<img';
	if (id) html += ' id="'+id+'"';
	html += ' src="' + this.src + '">';
	return html;
}



